"""llama.cpp server adapter.

See more details:
https://github.com/ggerganov/llama.cpp/blob/master/examples/server/README.md

**Features:**
 * LLM inference of F16 and quantized models on GPU and CPU
 * Parallel decoding with multi-user support
 * Continuous batching

The llama.cpp server is pure C++ server, we need to use the llama-cpp-server-py-core
to interact with it.
"""

import logging
from dataclasses import dataclass, field, fields
from typing import List, Optional, Type

from dbgpt.core.interface.parameter import LLMDeployModelParameters
from dbgpt.model.adapter.base import (
    ConversationAdapter,
    LLMModelAdapter,
    register_model_adapter,
)
from dbgpt.model.adapter.model_metadata import COMMON_LLAMA_CPP_MODELS
from dbgpt.model.base import ModelType
from dbgpt.util.i18n_utils import _

logger = logging.getLogger(__name__)


_DEFAULT_HF_REPO = "ggml-org/models"
_DEFAULT_HF_FILE = "tinyllamas/stories260K.gguf"
_DEFAULT_MODEL_ALIAS = "tinyllama-2"


@dataclass
class LlamaServerParameters(LLMDeployModelParameters):
    provider: str = ModelType.LLAMA_CPP_SERVER

    path: Optional[str] = field(
        default=None,
        metadata={
            "order": -800,
            "help": _("Local model file path"),
        },
    )
    # Model configuration
    model_hf_repo: Optional[str] = field(
        default=None,
        metadata={"help": _("Hugging Face repository for model download")},
    )

    model_hf_file: Optional[str] = field(
        default=None,
        metadata={"help": _("Model file name in the Hugging Face repository")},
    )

    device: Optional[str] = field(
        default=None,
        metadata={
            "order": -700,
            "help": _(
                "Device to run model. If None, the device is automatically determined"
            ),
        },
    )

    # Server configuration
    server_bin_path: Optional[str] = field(
        default=None, metadata={"help": _("Path to the server binary executable")}
    )

    server_host: str = field(
        default="127.0.0.1", metadata={"help": _("Host address to bind the server")}
    )

    server_port: int = field(
        default=0,
        metadata={"help": _("Port to bind the server. 0 for random available port")},
    )

    # Model parameters
    temperature: float = field(
        default=0.8, metadata={"help": _("Sampling temperature for text generation")}
    )

    seed: int = field(
        default=42, metadata={"help": _("Random seed for reproducibility")}
    )

    debug: bool = field(default=False, metadata={"help": _("Enable debug mode")})

    # Optional model sources
    model_url: Optional[str] = field(
        default=None,
        metadata={"help": _("Model download URL (env: LLAMA_ARG_MODEL_URL)")},
    )

    model_draft: Optional[str] = field(
        default=None, metadata={"help": _("Draft model file path")}
    )

    # Performance settings
    threads: Optional[int] = field(
        default=None,
        metadata={
            "help": _(
                "Number of threads to use during generation (default: -1) "
                "(env: LLAMA_ARG_THREADS)"
            )
        },
    )

    n_gpu_layers: Optional[int] = field(
        default=None,
        metadata={
            "help": _(
                "Number of layers to store in VRAM (env: LLAMA_ARG_N_GPU_LAYERS), set"
                " 1000000000 to use all layers"
            )
        },
    )

    batch_size: Optional[int] = field(
        default=None,
        metadata={
            "help": _(
                "Logical maximum batch size (default: 2048) (env: LLAMA_ARG_BATCH)"
            )
        },
    )

    ubatch_size: Optional[int] = field(
        default=None,
        metadata={
            "help": _(
                "Physical maximum batch size (default: 512) (env: LLAMA_ARG_UBATCH)"
            )
        },
    )

    # Context settings
    ctx_size: Optional[int] = field(
        default=None,
        metadata={
            "help": _(
                "Size of the prompt context (default: 4096, 0 = loaded from model) "
                "(env: LLAMA_ARG_CTX_SIZE)"
            )
        },
    )

    grp_attn_n: Optional[int] = field(
        default=None, metadata={"help": _("Group-attention factor (default: 1)")}
    )

    grp_attn_w: Optional[int] = field(
        default=None, metadata={"help": _("Group-attention width (default: 512)")}
    )

    # Generation settings
    n_predict: Optional[int] = field(
        default=None,
        metadata={
            "help": _(
                "Number of tokens to predict (default: -1, -1 = infinity, -2 = until "
                "context filled) (env: LLAMA_ARG_N_PREDICT)"
            )
        },
    )

    # Cache settings
    slot_save_path: Optional[str] = field(
        default=None,
        metadata={"help": _("Path to save slot kv cache (default: disabled)")},
    )

    n_slots: Optional[int] = field(
        default=None, metadata={"help": _("Number of slots for KV cache")}
    )

    # Feature flags
    cont_batching: bool = field(
        default=False,
        metadata={"help": _("Enable continuous batching (a.k.a dynamic batching)")},
    )

    embedding: bool = field(
        default=False,
        metadata={
            "help": _(
                "Restrict to only support embedding use case; use only with dedicated "
                "embedding models (env: LLAMA_ARG_EMBEDDINGS)"
            )
        },
    )

    reranking: bool = field(
        default=False,
        metadata={
            "help": _("Enable reranking endpoint on server (env: LLAMA_ARG_RERANKING)")
        },
    )

    metrics: bool = field(
        default=False,
        metadata={
            "help": _(
                "Enable prometheus compatible metrics endpoint "
                "(env: LLAMA_ARG_ENDPOINT_METRICS)"
            )
        },
    )

    slots: bool = field(
        default=False,
        metadata={
            "help": _(
                "Enable slots monitoring endpoint (env: LLAMA_ARG_ENDPOINT_SLOTS)"
            )
        },
    )

    # Draft settings
    draft: Optional[int] = field(
        default=None,
        metadata={
            "help": _(
                "Number of tokens to draft for speculative decoding (default: 16) "
                "(env: LLAMA_ARG_DRAFT_MAX)"
            )
        },
    )

    draft_max: Optional[int] = field(
        default=None, metadata={"help": _("Same as draft")}
    )

    draft_min: Optional[int] = field(
        default=None,
        metadata={
            "help": _(
                "Minimum number of draft tokens to use for speculative decoding "
                "(default: 5)"
            )
        },
    )

    # Authentication
    api_key: Optional[str] = field(
        default=None,
        metadata={"help": _("API key to use for authentication (env: LLAMA_API_KEY)")},
    )

    # Model adaptation
    lora_files: List[str] = field(
        default_factory=list,
        metadata={
            "help": _("Path to LoRA adapter (can be repeated to use multiple adapters)")
        },
    )

    # Additional settings
    no_context_shift: bool = field(
        default=False,
        metadata={"help": _("Disables context shift on infinite text generation")},
    )

    no_webui: Optional[bool] = field(
        default=None, metadata={"help": _("Disable web UI")}
    )
    startup_timeout: Optional[int] = field(
        default=None, metadata={"help": _("Server startup timeout in seconds")}
    )

    concurrency: Optional[int] = field(
        default=20, metadata={"help": _("Model concurrency limit")}
    )

    @property
    def real_model_path(self) -> Optional[str]:
        """Get the real model path.

        If deploy model is not local, return None.
        """
        return self._resolve_root_path(self.path)

    @property
    def real_device(self) -> Optional[str]:
        """Get the real device."""
        return self.device or super().real_device

    def generate_server_config(self):
        from llama_cpp_server_py_core import ServerConfig

        curr_config = self.to_dict()
        config_dict = {}
        for fd in fields(ServerConfig):
            if fd.name in curr_config:
                config_dict[fd.name] = curr_config[fd.name]

        if (
            self.real_device
            and self.real_device == "cuda"
            and ("n_gpu_layers" not in config_dict or not config_dict["n_gpu_layers"])
        ):
            # Set n_gpu_layers to a large number to use all layers
            logger.info("Set n_gpu_layers to a large number to use all layers")
            config_dict["n_gpu_layers"] = 1000000000
        config_dict["model_alias"] = self.name
        config_dict["model_file"] = self._resolve_root_path(self.path)
        model_file = config_dict.get("model_file")
        model_url = config_dict.get("model_url")
        model_hf_repo = config_dict.get("model_hf_repo")
        if not model_file and not model_url and not model_hf_repo:
            # Default to use Hugging Face repository
            config_dict["model_hf_repo"] = _DEFAULT_HF_REPO
            config_dict["model_hf_file"] = _DEFAULT_HF_FILE
        return ServerConfig(**config_dict)


class LLamaServerModelAdapter(LLMModelAdapter):
    def match(
        self,
        provider: str,
        model_name: Optional[str] = None,
        model_path: Optional[str] = None,
    ) -> bool:
        return provider == ModelType.LLAMA_CPP_SERVER

    def model_type(self) -> str:
        return ModelType.LLAMA_CPP_SERVER

    def model_param_class(self, model_type: str = None) -> Type[LlamaServerParameters]:
        return LlamaServerParameters

    def get_default_conv_template(
        self, model_name: str, model_path: str
    ) -> Optional[ConversationAdapter]:
        return None

    def load_from_params(self, params: LlamaServerParameters):
        """Load model from parameters."""
        try:
            from llama_cpp_server_py_core import LlamaCppServer, ServerProcess
        except ImportError:
            logger.error(
                "Failed to import llama_cpp_server_py_core, please install it first by "
                "`pip install llama-cpp-server-py`"
            )
            raise

        server_config = params.generate_server_config()
        server = ServerProcess(server_config)
        server.start(params.startup_timeout or 300)
        model_server = LlamaCppServer(server, server_config)
        return model_server, model_server

    def support_generate_function(self) -> bool:
        return True

    def get_generate_stream_function(
        self, model, deploy_model_params: LLMDeployModelParameters
    ):
        from dbgpt.model.llm.llama_cpp.llama_cpp_server import generate_stream

        return generate_stream

    def get_generate_function(
        self, model, deploy_model_params: LLMDeployModelParameters
    ):
        from dbgpt.model.llm.llama_cpp.llama_cpp_server import generate

        return generate


register_model_adapter(
    LLamaServerModelAdapter, supported_models=COMMON_LLAMA_CPP_MODELS
)



import hashlib
import json
import logging
from dataclasses import is_dataclass
from pathlib import Path
from typing import Any, Dict, List, Set, Tuple, Type

from dbgpt.util.configure.manager import (
    ConfigurationManager,
    RegisterParameters,
)
from dbgpt.util.module_utils import import_from_string
from dbgpt.util.parameter_utils import ParameterDescription

logger = logging.getLogger(__name__)


def is_direct_subclass(cls, parent_cls):
    """Determine if a class is a direct subclass of another class."""
    # Check if the class is a subclass of the parent class
    if not issubclass(cls, parent_cls):
        return False

    # Check if the class is a direct subclass of the parent class
    return parent_cls in cls.__bases__


def _is_skip_class(cls):
    return is_direct_subclass(cls, RegisterParameters)


class MDXDocGenerator:
    def __init__(self):
        self.processed_classes: Set[str] = set()
        self.link_cache: Dict[str, str] = {}  # {class_path: filename}
        self._doc_cache: Dict[str, str] = {}
        self.generated_files: Set[str] = set()

    def generate_safe_filename(self, doc_id: str) -> str:
        """Generate a safe filename for the given doc_id."""
        if doc_id in self.link_cache:
            return self.link_cache[doc_id]

        parts = doc_id.split(".")[-2:]  # Just use the last two parts
        main_part = "_".join(parts)
        hash_suffix = hashlib.md5(doc_id.encode()).hexdigest()[:6]
        filename = f"{main_part}_{hash_suffix}.mdx".lower()
        self.link_cache[doc_id] = filename
        return filename

    def get_abs_link(self, cls: Type, doc_id: str) -> str:
        filename = self.generate_safe_filename(doc_id)
        link_url = "/docs/config-reference/"
        cfg_type, _ = self._parse_class_metadata(cls)
        if cfg_type:
            link_url += f"{cfg_type}/"
        link_url += filename[:-4]
        return link_url

    def get_rel_link(self, cls: Type, doc_id: str, source_cls: Type = None) -> str:
        """Generate a relative link from the source class to the target class.

        Args:
            cls: The target class to generate a link to
            doc_id: The document ID of the target class
            source_cls: The source class that will contain the link (optional)

        Returns:
            A relative URL path to the target class documentation
        """
        filename = self.generate_safe_filename(doc_id)
        target_type, _ = self._parse_class_metadata(cls)

        # If source_cls is not provided, return a simple path without relative
        # navigation
        if not source_cls:
            if target_type:
                return f"{target_type}/{filename[:-4]}"
            return filename[:-4]

        # Get the source class type to determine relative path
        source_type, _ = self._parse_class_metadata(source_cls)

        # Same type - link within the same directory
        if source_type == target_type:
            return filename[:-4]

        # Different types - need to navigate up and then down
        if source_type and target_type:
            return f"../{target_type}/{filename[:-4]}"
        elif source_type and not target_type:
            return f"../{filename[:-4]}"
        elif not source_type and target_type:
            return f"{target_type}/{filename[:-4]}"
        else:
            return filename[:-4]

    def get_desc_for_class(self, cls: Type, default_desc: str = "") -> str:
        """Get the description for a class."""
        doc_id = self.get_class_doc_id(cls)
        return self._doc_cache.get(doc_id, default_desc)

    def get_class_doc_id(self, cls: Type) -> str:
        """Get the unique identifier for a class."""
        return f"{cls.__module__}.{cls.__name__}"

    def convert_to_mdx_dict(self, value: Any) -> Dict:
        """Convert a value to a dictionary that can be used in MDX components."""
        if value is None:
            return {"type": "code", "content": "None"}

        if is_dataclass(type(value)):
            return {"type": "code", "content": value.__class__.__name__}

        if isinstance(value, str):
            if "${" in value or "/" in value:
                return {"type": "code", "content": value}
            if len(value) > 50:
                return {"type": "codeblock", "content": value}
            return {"type": "code", "content": value}

        if isinstance(value, (list, dict, tuple)):
            return {"type": "codeblock", "language": "python", "content": repr(value)}

        return {"type": "code", "content": str(value)}

    def process_nested_fields(
        self,
        nested_fields: Dict[str, List[ParameterDescription]],
        output_dir: Path,
        source_cls: Type,
    ) -> Tuple[List[Dict], List[str]]:
        """Handle nested fields in a parameter description."""
        links = []
        generated_files = []

        for type_name, params in nested_fields.items():
            if not params:
                continue

            try:
                nested_cls = import_from_string(params[0].param_class)

                doc_id = self.get_class_doc_id(nested_cls)

                if doc_id not in self.processed_classes:
                    new_files = self.generate_class_doc(nested_cls, output_dir)
                    generated_files.extend(new_files)
                # Use relative link instead of absolute link
                link_url = self.get_rel_link(nested_cls, doc_id, source_cls=source_cls)
                links.append(
                    {
                        "type": "link",
                        "text": f"{type_name} configuration",
                        "url": f"{link_url}",
                    }
                )

            except ImportError:
                logger.warning(
                    f"Cann't import configuration class: {params[0].param_class}"
                )
                links.append({"type": "text", "content": type_name})

        return links, generated_files

    def _parse_class_doc(self, param_cls: Type) -> Tuple[str, str, str]:
        metadata_name = f"_resource_metadata_{param_cls.__name__}"
        if hasattr(param_cls, metadata_name):
            from dbgpt.core.awel.flow import ResourceMetadata

            flow_metadata: ResourceMetadata = getattr(param_cls, metadata_name)
            label = flow_metadata.label
            description = flow_metadata.description
            documentation_url = flow_metadata.documentation_url
        else:
            label = param_cls.__name__
            description = param_cls.__doc__.strip() if param_cls.__doc__ else ""
            documentation_url = ""
        return label, description, documentation_url

    def _parse_class_metadata(self, param_cls: Type) -> Tuple[str | None, str | None]:
        cfg_type = None
        cfg_desc = None
        if hasattr(param_cls, "__cfg_type__"):
            cfg_type = getattr(param_cls, "__cfg_type__")
        if hasattr(param_cls, "__cfg_desc__"):
            cfg_desc = getattr(param_cls, "__cfg_desc__")
        return cfg_type, cfg_desc

    def generate_class_doc(self, cls: Type, output_dir: Path) -> List[str]:
        """Generate documentation for a configuration class."""
        # if _is_base_config(cls) or cls is RegisterParameters:
        #     return []
        if not is_dataclass(cls):
            return []
        doc_id = self.get_class_doc_id(cls)
        if doc_id in self.processed_classes:
            return []

        self.processed_classes.add(doc_id)
        generated_files = []

        descriptions = ConfigurationManager.parse_description(
            cls, cache_enable=True, verbose=True
        )
        cfg_type, cfg_desc = self._parse_class_metadata(cls)

        filename = self.generate_safe_filename(doc_id)
        if cfg_type:
            output_path = output_dir / cfg_type / filename
        else:
            output_path = output_dir / filename

        if not output_path.parent.exists():
            output_path.parent.mkdir(parents=True, exist_ok=True)
        generated_files.append(filename)

        cls_label, cls_desc, cls_doc_url = self._parse_class_doc(cls)
        self._doc_cache[doc_id] = cls_desc
        with open(output_path, "w", encoding="utf-8") as f:
            # Add frontmatter to set the page title
            f.write("---\n")
            f.write(f'title: "{cls_label} Configuration"\n')
            f.write(f'description: "{cls_desc}"\n')
            f.write("---\n\n")

            # Use custom component to render the config detail
            _config_component = '"@site/src/components/mdx/ConfigDetail";\n\n'
            f.write("import { ConfigDetail } from " + _config_component)

            # Build the full config data
            config_data = {
                "name": cls.__name__,
                "description": cls_desc,
                "documentationUrl": cls_doc_url,
                "parameters": [],
            }

            for param in sorted(
                descriptions,
                key=lambda x: (
                    x.param_order
                    if x.param_order is not None
                    else (1000 if x.required else float("inf"))
                ),
            ):
                param_data = {
                    "name": param.param_name,
                    "type": param.param_type or "",
                    "required": param.required,
                    "description": param.description or "",
                }
                # Handle nested fields
                if param.nested_fields:
                    nested_links, nested_files = self.process_nested_fields(
                        param.nested_fields, output_dir, source_cls=cls
                    )
                    generated_files.extend(nested_files)
                    if nested_links:
                        param_data["nestedTypes"] = nested_links

                # Handle default value
                if param.default_value is not None:
                    if is_dataclass(type(param.default_value)):
                        param_data["defaultValue"] = (
                            param.default_value.__class__.__name__
                        )
                    else:
                        param_data["defaultValue"] = str(param.default_value)

                # Handle valid values
                if param.valid_values:
                    param_data["validValues"] = [str(v) for v in param.valid_values]

                config_data["parameters"].append(param_data)

            # Write the config detail component
            f.write("<ConfigDetail config={")
            f.write(json.dumps(config_data, indent=2, ensure_ascii=False))
            f.write("} />\n\n")

            # If the class is abstract, list all concrete implementations
            if hasattr(cls, "__abstract__"):
                subs = [c for c in cls.__subclasses__() if is_dataclass(c)]
                if subs:
                    f.write("## Implementations\n\n")
                    for sub_cls in subs:
                        sub_files = self.generate_class_doc(sub_cls, output_dir)
                        generated_files.extend(sub_files)

                        sub_doc_id = self.get_class_doc_id(sub_cls)
                        sub_filename = self.generate_safe_filename(sub_doc_id)
                        f.write(
                            f"- [{sub_cls.__name__} configuretion](./{sub_filename})\n"
                        )

        self.generated_files.add(filename)
        return generated_files

    def _collect_relationships(
        self, cls: Type, processed: Set[str], relationships: List[Dict]
    ):
        """Collect relationships between classes."""

        if not is_dataclass(cls):
            return
        class_id = self.get_class_doc_id(cls)
        if class_id in processed:
            return
        processed.add(class_id)

        descriptions = ConfigurationManager.parse_description(
            cls, cache_enable=True, verbose=True
        )
        for param in descriptions:
            if param.nested_fields:
                for nested_type, nested_params in param.nested_fields.items():
                    if nested_params:
                        try:
                            nested_cls = import_from_string(
                                nested_params[0].param_class
                            )
                            relationships.append(
                                {
                                    "from": cls.__name__,
                                    "to": nested_cls.__name__,
                                    "label": param.param_name,
                                }
                            )
                            self._collect_relationships(
                                nested_cls, processed, relationships
                            )
                        except ImportError:
                            logger.warning(
                                "Can't import configuration class: "
                                f"{nested_params[0].param_class}"
                            )

    def generate_overview(self, output_dir: Path, config_classes: List[Type]):
        """Generate an overview document for all configuration classes grouped by
        __cfg_type__."""

        overview_path = output_dir / "overview.mdx"

        # Collect all relationships between classes
        relationships = []
        processed = set()

        for cls in config_classes:
            self._collect_relationships(cls, processed, relationships)

        # Group classes by __cfg_type__
        type_groups = {}
        class_map = {
            cls.__name__: cls for cls in config_classes if hasattr(cls, "__name__")
        }

        # Get all classes and group them by __cfg_type__
        for cls in config_classes:
            if _is_skip_class(cls):
                logger.info("Skipping class: %s", cls)
                continue
            # Otherwise, default to "other"
            cfg_type = getattr(cls, "__cfg_type__", "other")
            if not cfg_type:
                cfg_type = "other"

            if cfg_type not in type_groups:
                type_groups[cfg_type] = []
            type_groups[cfg_type].append(cls)

        # type_relationships = {cfg_type: [] for cfg_type in type_groups}
        type_relationships = {}
        for rel in relationships:
            from_cls = class_map.get(rel["from"])
            to_cls = class_map.get(rel["to"])

            if from_cls and to_cls:
                from_type = getattr(from_cls, "__cfg_type__", "other") or "other"
                to_type = getattr(to_cls, "__cfg_type__", "other") or "other"

                # If the two classes belong to the same type, add the relationship to
                # that type
                if from_type == to_type:
                    if from_type not in type_relationships:
                        type_relationships[from_type] = []
                    type_relationships[from_type].append(rel)

        # Write the overview document
        with open(overview_path, "w", encoding="utf-8") as f:
            f.write("---\n")
            f.write('title: "Configuration Overview"\n')
            f.write("---\n\n")

            f.write("# Configuration Overview\n\n")
            f.write(
                "This document provides an overview of all configuration classes "
                "organized by type.\n\n"
            )

            # Create a table of contents with links to each type
            f.write("## Configuration Types\n\n")

            for cfg_type, classes in sorted(type_groups.items()):
                f.write(
                    f"- [{cfg_type}](#type-{cfg_type.lower().replace(' ', '-')}) ({len(classes)} classes)\n"  # noqa
                )

            f.write("\n## Type Details\n\n")

            # Create a section for each type
            for cfg_type, classes in sorted(type_groups.items()):
                if cfg_type != "other":
                    index_path = output_dir / f"{cfg_type}/index.mdx"
                    with open(index_path, "w", encoding="utf-8") as f_index:
                        f_index.write("---\n")
                        f_index.write(f'title: "{cfg_type}"\n')
                        f_index.write(f'description: "{cfg_type} Configuration"\n')
                        f_index.write("---\n\n")
                        f_index.write(
                            f"# {cfg_type} Configuration\n\n"
                            f"This document provides an overview of all configuration classes in {cfg_type} type.\n\n"  # noqa
                        )

                        # Use custom component to render the config class table
                        f_index.write(
                            "import { ConfigClassTable } from '@site/src/components/mdx/ConfigClassTable';\n\n"  # noqa
                        )  # noqa

                        # Write the config class table component
                        f_index.write("## Configuration Classes\n\n")
                        f_index.write("<ConfigClassTable classes={[\n")

                        # Write the config class table component
                        for cls in sorted(classes, key=lambda x: x.__name__):
                            cls_name = cls.__name__
                            cfg_desc = getattr(cls, "__cfg_desc__", "") or ""
                            cfg_desc = self.get_desc_for_class(cls, cfg_desc)

                            doc_id = self.get_class_doc_id(cls)
                            # doc_link = self.get_abs_link(cls, doc_id)
                            doc_link = self.generate_safe_filename(doc_id)
                            f_index.write("  {\n")
                            f_index.write(f'    "name": "{cls_name}",\n')
                            f_index.write(
                                f'    "description": {json.dumps(cfg_desc)},\n'
                            )  # noqa
                            if doc_link:
                                f_index.write(f'    "link": "./{doc_link[:-4]}"\n')
                            else:
                                f_index.write('    "link": ""\n')
                            f_index.write("  },\n")

                        f_index.write("]} />\n\n")

                # Create an anchor by replacing spaces with dashes
                anchor = f"type-{cfg_type.lower().replace(' ', '-')}"
                f.write(f"### {cfg_type} {{#{anchor}}}\n\n")

                # Show the number of classes in this type
                f.write(f"This type contains {len(classes)} configuration classes.\n\n")

                # Obtain relationships for this type
                type_rels = type_relationships.get(cfg_type, [])

                # If the number of relationships is moderate, draw a relationship
                # diagram
                if 1 <= len(type_rels) <= 30:
                    f.write("#### Relationships\n\n")
                    f.write("```mermaid\ngraph TD\n")

                    for rel in type_rels:
                        f.write(f"    {rel['from']} -->|{rel['label']}| {rel['to']}\n")

                    f.write("```\n\n")
                elif len(type_rels) > 30:
                    # If there are too many relationships, add a note
                    f.write(
                        f"This type has {len(type_rels)} relationships, which is too many to display in a single diagram.\n\n"  # noqa
                    )

                # List all classes and their links
                f.write("#### Configuration Classes\n\n")
                f.write("| Class | Description |\n")
                f.write("|-------|-------------|\n")

                for cls in sorted(classes, key=lambda x: x.__name__):
                    cls_name = cls.__name__
                    cfg_desc = getattr(cls, "__cfg_desc__", "")
                    # cfg_desc = self.get_desc_for_class(cls, cfg_desc)
                    # cfg_desc = cfg_desc.replace("`", "'")
                    doc_id = self.get_class_doc_id(cls)
                    if doc_id in self.link_cache:
                        # Use relative links based on the config type
                        if cfg_type != "other":
                            link_url = (
                                f"{cfg_type}/{self.generate_safe_filename(doc_id)[:-4]}"
                            )
                        else:
                            link_url = f"{self.generate_safe_filename(doc_id)[:-4]}"
                        f.write(f"| [{cls_name}]({link_url}) | {cfg_desc} |\n")
                    else:
                        f.write(f"| {cls_name} | {cfg_desc} |\n")

                f.write("\n---\n\n")  # Split sections

            # Add cross-type relationships
            f.write("## Cross-Type Relationships\n\n")
            f.write(
                "The following diagram shows relationships between different configuration types:\n\n"  # noqa
            )

            # Create relationships between types
            type_to_type = {}
            for rel in relationships:
                from_cls = class_map.get(rel["from"])
                to_cls = class_map.get(rel["to"])

                if from_cls and to_cls:
                    from_type = getattr(from_cls, "__cfg_type__", "other") or "other"
                    to_type = getattr(to_cls, "__cfg_type__", "other") or "other"

                    if from_type != to_type:
                        key = f"{from_type}|{to_type}"
                        if key not in type_to_type:
                            type_to_type[key] = 0
                        type_to_type[key] += 1

            if type_to_type:
                f.write("```mermaid\ngraph TD\n")

                # Add nodes for each type
                for cfg_type in type_groups:
                    f.write(
                        f"    {cfg_type}[{cfg_type} - {len(type_groups[cfg_type])} classes]\n"  # noqa
                    )

                # Add relationships between types
                for key, count in type_to_type.items():
                    from_type, to_type = key.split("|")
                    f.write(f"    {from_type} -->|{count} connections| {to_type}\n")

                f.write("```\n\n")
            else:
                f.write("No cross-type relationships found.\n\n")

            # Add a search prompt
            f.write("## Looking for a specific configuration?\n\n")
            f.write("1. Use the search function in the documentation site\n")
            f.write("2. Browse the configuration types above\n")
            f.write(
                "3. Check the specific class documentation for detailed parameter information\n"  # noqa
            )


def generate_docs(config_classes: List[Type], output_path: str):
    """The entry point for generating documentation."""
    generator = MDXDocGenerator()
    output_dir = Path(output_path)
    backup_dir = None
    output_dir.mkdir(parents=True, exist_ok=True)
    if output_dir.exists():
        backup_dir = Path("/tmp") / "dbgpt_tmp_backup"
        if backup_dir.exists():
            os.system(f"rm -rf {backup_dir}")
        # Backup the existing files, Move the existing files to the backup directory
        os.system(f"mv {output_dir} {backup_dir}")

    output_dir.mkdir(parents=True, exist_ok=True)

    ConfigurationManager._description_cache = {}
    # Generate documentation for each class
    try:
        all_generated_files = []
        for cls in config_classes[:1]:
            # Just generate the first class for now
            generated_files = generator.generate_class_doc(cls, output_dir)
            all_generated_files.extend(generated_files)

        # Generate an overview
        generator.generate_overview(output_dir, config_classes)
    except Exception as e:
        logger.error(f"Error generating documentation: {e}")
        # Restore the backup
        if backup_dir:
            # Delete the output directory
            os.system(f"rm -rf {output_dir}")
            os.system(f"mv {backup_dir} {output_dir}")
        raise e

    return all_generated_files


def _get_all_subclasses(cls):
    all_subclasses = []
    direct_subclasses = cls.__subclasses__()
    all_subclasses.extend(direct_subclasses)
    for subclass in direct_subclasses:
        all_subclasses.extend(_get_all_subclasses(subclass))

    return all_subclasses


if __name__ == "__main__":
    import os
    import sys

    logging.basicConfig(
        stream=sys.stdout,
        level=logging.INFO,
        format="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
    )

    from dbgpt.configs.model_config import ROOT_PATH
    from dbgpt_app.config import ApplicationConfig
    from dbgpt_app.dbgpt_server import scan_configs

    output_path = os.path.join(ROOT_PATH, "docs", "docs", "config-reference")

    scan_configs()

    config_classes = [ApplicationConfig]
    for subclass in _get_all_subclasses(RegisterParameters):
        config_classes.append(subclass)
    logger.info(f"Generating docs for {len(config_classes)} classes")
    logger.info(f"Generated for classes: {config_classes}")

    generate_docs(config_classes, output_path=output_path)



import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, Optional, Type, Union

from dbgpt.core import ModelMetadata
from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.model.proxy.llms.proxy_model import ProxyModel, parse_model_request
from dbgpt.util.i18n_utils import _

from ..base import (
    AsyncGenerateStreamFunction,
    GenerateStreamFunction,
    register_proxy_model_adapter,
)
from .chatgpt import OpenAICompatibleDeployModelParameters, OpenAILLMClient

if TYPE_CHECKING:
    from httpx._types import ProxiesTypes
    from openai import AsyncAzureOpenAI, AsyncOpenAI

    ClientType = Union[AsyncAzureOpenAI, AsyncOpenAI]


_INFINIAI_DEFAULT_MODEL = "deepseek-v3"


@auto_register_resource(
    label=_("InfiniAI Proxy LLM"),
    category=ResourceCategory.LLM_CLIENT,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("InfiniAI proxy LLM configuration."),
    documentation_url="https://docs.infini-ai.com/gen-studio/api/tutorial.html",  # noqa
    show_in_ui=False,
)
@dataclass
class InfiniAIDeployModelParameters(OpenAICompatibleDeployModelParameters):
    """Deploy model parameters for InfiniAI."""

    provider: str = "proxy/infiniai"

    api_base: Optional[str] = field(
        default="${env:INFINIAI_API_BASE:-https://cloud.infini-ai.com/maas/v1}",
        metadata={
            "help": _("The base url of the InfiniAI API."),
        },
    )

    api_key: Optional[str] = field(
        default="${env:INFINIAI_API_KEY}",
        metadata={
            "help": _("The API key of the InfiniAI API."),
            "tags": "privacy",
        },
    )


async def infiniai_generate_stream(
    model: ProxyModel, tokenizer, params, device, context_len=2048
):
    client: InfiniAILLMClient = model.proxy_llm_client
    request = parse_model_request(params, client.default_model, stream=True)
    async for r in client.generate_stream(request):
        yield r


class InfiniAILLMClient(OpenAILLMClient):
    """InfiniAI LLM Client.

    InfiniAI's API is compatible with OpenAI's API, so we inherit from
    OpenAILLMClient.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        api_type: Optional[str] = None,
        api_version: Optional[str] = None,
        model: Optional[str] = _INFINIAI_DEFAULT_MODEL,
        proxies: Optional["ProxiesTypes"] = None,
        timeout: Optional[int] = 240,
        model_alias: Optional[str] = _INFINIAI_DEFAULT_MODEL,
        context_length: Optional[int] = None,
        openai_client: Optional["ClientType"] = None,
        openai_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        api_base = (
            api_base
            or os.getenv("INFINIAI_API_BASE")
            or "https://cloud.infini-ai.com/maas/v1"
        )
        api_key = api_key or os.getenv("INFINIAI_API_KEY")
        model = model or _INFINIAI_DEFAULT_MODEL
        if not context_length:
            if "200k" in model:
                context_length = 200 * 1024
            else:
                context_length = 4096

        if not api_key:
            raise ValueError(
                "InfiniAI API key is required, please set 'INFINIAI_API_KEY' in "
                "environment or pass it as an argument."
            )

        super().__init__(
            api_key=api_key,
            api_base=api_base,
            api_type=api_type,
            api_version=api_version,
            model=model,
            proxies=proxies,
            timeout=timeout,
            model_alias=model_alias,
            context_length=context_length,
            openai_client=openai_client,
            openai_kwargs=openai_kwargs,
            **kwargs,
        )

    @property
    def default_model(self) -> str:
        model = self._model
        if not model:
            model = _INFINIAI_DEFAULT_MODEL
        return model

    @classmethod
    def param_class(cls) -> Type[InfiniAIDeployModelParameters]:
        return InfiniAIDeployModelParameters

    @classmethod
    def generate_stream_function(
        cls,
    ) -> Optional[Union[GenerateStreamFunction, AsyncGenerateStreamFunction]]:
        return infiniai_generate_stream


register_proxy_model_adapter(
    InfiniAILLMClient,
    supported_models=[
        ModelMetadata(
            model=["deepseek-v3"],
            context_length=64 * 1024,
            max_output_length=8 * 1024,
            description="DeepSeek-V3 by DeepSeek",
            link="https://cloud.infini-ai.com/genstudio/model",
            function_calling=True,
        ),
        ModelMetadata(
            model=["deepseek-r1"],
            context_length=64 * 1024,
            max_output_length=8 * 1024,
            description="DeepSeek-V3 by DeepSeek",
            link="https://cloud.infini-ai.com/genstudio/model",
            function_calling=False,
        ),
        ModelMetadata(
            model=["qwq-32b"],
            context_length=64 * 1024,
            max_output_length=8 * 1024,
            description="qwq By Qwen",
            link="https://cloud.infini-ai.com/genstudio/model",
            function_calling=True,
        ),
        ModelMetadata(
            model=[
                "qwen2.5-72b-instruct",
                "qwen2.5-32b-instruct",
                "qwen2.5-14b-instruct",
                "qwen2.5-7b-instruct",
                "qwen2.5-coder-32b-instruct",
            ],
            context_length=32 * 1024,
            max_output_length=4 * 1024,
            description="Qwen 2.5 By Qwen",
            link="https://cloud.infini-ai.com/genstudio/model",
            function_calling=True,
        ),
        # More models see: https://cloud.infiniai.cn/models
    ],
)



import json
import os
from concurrent.futures import Executor
from dataclasses import dataclass, field
from typing import Iterator, Optional, Type, Union

from dbgpt.core import MessageConverter, ModelMetadata, ModelOutput, ModelRequest
from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.model.proxy.base import (
    AsyncGenerateStreamFunction,
    GenerateStreamFunction,
    ProxyLLMClient,
    register_proxy_model_adapter,
)
from dbgpt.model.proxy.llms.proxy_model import ProxyModel, parse_model_request
from dbgpt.util.i18n_utils import _

from .chatgpt import OpenAICompatibleDeployModelParameters


@auto_register_resource(
    label=_("Xunfei Spark Proxy LLM"),
    category=ResourceCategory.LLM_CLIENT,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("Xunfei Spark proxy LLM configuration."),
    documentation_url="https://www.xfyun.cn/doc/spark/HTTP%E8%B0%83%E7%94%A8%E6%96%87%E6%A1%A3.html#_1-%E6%8E%A5%E5%8F%A3%E8%AF%B4%E6%98%8E",
    show_in_ui=False,
)
@dataclass
class SparkDeployModelParameters(OpenAICompatibleDeployModelParameters):
    """Deploy model parameters for Spark."""

    provider: str = "proxy/spark"

    api_base: Optional[str] = field(
        default="${env:XUNFEI_SPARK_API_BASE:-https://spark-api-open.xf-yun.com/v1}",
        metadata={
            "help": _("The base url of the Spark API."),
        },
    )

    api_key: Optional[str] = field(
        default="${env:XUNFEI_SPARK_API_KEY}",
        metadata={
            "help": _("The API key of the Spark API."),
            "tags": "privacy",
        },
    )


def spark_generate_stream(
    model: ProxyModel, tokenizer, params, device, context_len=2048
):
    client: SparkLLMClient = model.proxy_llm_client
    request = parse_model_request(params, client.default_model, stream=True)
    for r in client.sync_generate_stream(request):
        yield r


def extract_content(line: str):
    if not line.strip():
        return line
    if line.startswith("data: "):
        json_str = line[len("data: ") :]
    else:
        raise ValueError("Error line content ")

    try:
        data = json.loads(json_str)
        if data == "[DONE]":
            return ""

        choices = data.get("choices", [])
        if choices and isinstance(choices, list):
            delta = choices[0].get("delta", {})
            content = delta.get("content", "")
            return content
        else:
            raise ValueError("Error line content ")
    except json.JSONDecodeError:
        return ""


class SparkLLMClient(ProxyLLMClient):
    def __init__(
        self,
        model: Optional[str] = None,
        api_base: Optional[str] = None,
        api_key: Optional[str] = None,
        model_alias: Optional[str] = "spark_proxyllm",
        context_length: Optional[int] = 4096,
        executor: Optional[Executor] = None,
    ):
        """
        星火大模型API当前有Lite、Pro、Pro-128K、Max、Max-32K和4.0 Ultra六个版本
        Spark4.0 Ultra 请求地址，对应的domain参数为4.0Ultra
        Spark Max-32K请求地址，对应的domain参数为max-32k
        Spark Max请求地址，对应的domain参数为generalv3.5
        Spark Pro-128K请求地址，对应的domain参数为pro-128k：
        Spark Pro请求地址，对应的domain参数为generalv3：
        Spark Lite请求地址，对应的domain参数为lite：
        https://www.xfyun.cn/doc/spark/HTTP%E8%B0%83%E7%94%A8%E6%96%87%E6%A1%A3.html#_3-%E8%AF%B7%E6%B1%82%E8%AF%B4%E6%98%8E
        """
        self._model = model or os.getenv("XUNFEI_SPARK_API_MODEL")
        self._api_base = self._resolve_env_vars(api_base)
        self._api_key = self._resolve_env_vars(
            api_key
            or os.getenv("XUNFEI_SPARK_API_KEY")
            or os.getenv("XUNFEI_SPARK_API_PASSWORD")
        )
        if not self._model:
            raise ValueError("model can't be empty")
        if not self._api_base:
            raise ValueError("api_base can't be empty")
        if not self._api_key:
            raise ValueError("Spark API key is required, please provide it.")

        super().__init__(
            model_names=[model, model_alias],
            context_length=context_length,
            executor=executor,
        )

    @classmethod
    def new_client(
        cls,
        model_params: SparkDeployModelParameters,
        default_executor: Optional[Executor] = None,
    ) -> "SparkLLMClient":
        return cls(
            model=model_params.real_provider_model_name,
            api_base=model_params.api_base,
            api_key=model_params.api_key,
            model_alias=model_params.real_provider_model_name,
            context_length=model_params.context_length,
            executor=default_executor,
        )

    @classmethod
    def param_class(cls) -> Type[SparkDeployModelParameters]:
        return SparkDeployModelParameters

    @classmethod
    def generate_stream_function(
        cls,
    ) -> Optional[Union[GenerateStreamFunction, AsyncGenerateStreamFunction]]:
        return spark_generate_stream

    @property
    def default_model(self) -> str:
        return self._model

    def sync_generate_stream(
        self,
        request: ModelRequest,
        message_converter: Optional[MessageConverter] = None,
    ) -> Iterator[ModelOutput]:
        """
        reference:
        https://www.xfyun.cn/doc/spark/HTTP%E8%B0%83%E7%94%A8%E6%96%87%E6%A1%A3.html#_3-%E8%AF%B7%E6%B1%82%E8%AF%B4%E6%98%8E
        """
        request = self.local_covert_message(request, message_converter)
        messages = request.to_common_messages(support_system_role=False)
        try:
            import requests
        except ImportError as e:
            raise ValueError(
                "Could not import python package: requests "
                "Please install requests by command `pip install requests"
            ) from e

        data = {
            "model": self._model,  # 指定请求的模型
            "messages": messages,
            "temperature": request.temperature,
            "stream": True,
        }
        header = {
            # Please replace the APIPassword with your own
            "Authorization": f"Bearer {self._api_key}"
        }
        url = self._api_base
        if url.endswith("/"):
            url = url[:-1]
        url = f"{url}/chat/completions"
        response = requests.post(url, headers=header, json=data, stream=True)
        response.encoding = "utf-8"
        try:
            content = ""
            # data: {"code":0,"message":"Success","sid":"cha000bf865@dx19307263c06b894532","id":"cha000bf865@dx19307263c06b894532","created":1730991766,"choices":[{"delta":{"role":"assistant","content":"你好"},"index":0}]} #noqa
            # data: [DONE]
            for line in response.iter_lines(decode_unicode=True):
                print("llm out:", line)
                content = content + extract_content(line)
                yield ModelOutput(text=content, error_code=0)
        except Exception as e:
            yield ModelOutput(
                text=f"**LLMServer Generate Error, Please CheckErrorInfo.**: {e}",
                error_code=1,
            )


register_proxy_model_adapter(
    SparkLLMClient,
    supported_models=[
        ModelMetadata(
            model="lite",
            # Maybe not correct, the value inferred from max_output_length
            context_length=14096,
            max_output_length=14096,
            description="Xunfei Spark Lite model",
            link="https://www.xfyun.cn/doc/spark/HTTP%E8%B0%83%E7%94%A8%E6%96%87%E6%A1%A3.html#_1-%E6%8E%A5%E5%8F%A3%E8%AF%B4%E6%98%8E",  # noqa
            function_calling=True,
        ),
        ModelMetadata(
            model="pro-128k",
            context_length=128 * 1024,
            max_output_length=14096,
            description="Xunfei Spark Pro-128K model",
            link="https://www.xfyun.cn/doc/spark/HTTP%E8%B0%83%E7%94%A8%E6%96%87%E6%A1%A3.html#_1-%E6%8E%A5%E5%8F%A3%E8%AF%B4%E6%98%8E",  # noqa
            function_calling=True,
        ),
        ModelMetadata(
            model=["generalv3", "generalv3.5", "4.0Ultra"],
            # Maybe not correct, the value inferred from max_output_length
            context_length=18192,
            max_output_length=18192,
            description="Xunfei Spark General model",
            link="https://www.xfyun.cn/doc/spark/HTTP%E8%B0%83%E7%94%A8%E6%96%87%E6%A1%A3.html#_1-%E6%8E%A5%E5%8F%A3%E8%AF%B4%E6%98%8E",  # noqa
            function_calling=True,
        ),
        ModelMetadata(
            model=["pro-128k"],
            context_length=128 * 1024,
            max_output_length=18192,
            description="Xunfei Spark Pro-128K model",
            link="https://www.xfyun.cn/doc/spark/HTTP%E8%B0%83%E7%94%A8%E6%96%87%E6%A1%A3.html#_1-%E6%8E%A5%E5%8F%A3%E8%AF%B4%E6%98%8E",  # noqa
            function_calling=True,
        ),
        ModelMetadata(
            model="max-32k",
            context_length=32 * 1024,
            max_output_length=18192,
            description="Xunfei Spark Max-32K model",
            link="https://www.xfyun.cn/doc/spark/HTTP%E8%B0%83%E7%94%A8%E6%96%87%E6%A1%A3.html#_1-%E6%8E%A5%E5%8F%A3%E8%AF%B4%E6%98%8E",  # noqa
            function_calling=True,
        ),
    ],
)



"""Proxy models."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dbgpt.model.proxy.llms.aimlapi import AimlapiLLMClient
    from dbgpt.model.proxy.llms.burncloud import BurnCloudLLMClient
    from dbgpt.model.proxy.llms.chatgpt import OpenAILLMClient
    from dbgpt.model.proxy.llms.claude import ClaudeLLMClient
    from dbgpt.model.proxy.llms.deepseek import DeepseekLLMClient
    from dbgpt.model.proxy.llms.gemini import GeminiLLMClient
    from dbgpt.model.proxy.llms.gitee import GiteeLLMClient
    from dbgpt.model.proxy.llms.infiniai import InfiniAILLMClient
    from dbgpt.model.proxy.llms.minimax import MiniMaxLLMClient
    from dbgpt.model.proxy.llms.moonshot import MoonshotLLMClient
    from dbgpt.model.proxy.llms.ollama import OllamaLLMClient
    from dbgpt.model.proxy.llms.siliconflow import SiliconFlowLLMClient
    from dbgpt.model.proxy.llms.spark import SparkLLMClient
    from dbgpt.model.proxy.llms.tongyi import TongyiLLMClient
    from dbgpt.model.proxy.llms.wenxin import WenxinLLMClient
    from dbgpt.model.proxy.llms.yi import YiLLMClient
    from dbgpt.model.proxy.llms.zhipu import ZhipuLLMClient


def __lazy_import(name):
    module_path = {
        "OpenAILLMClient": "dbgpt.model.proxy.llms.chatgpt",
        "BurnCloudLLMClient": "dbgpt.model.proxy.llms.burncloud",
        "ClaudeLLMClient": "dbgpt.model.proxy.llms.claude",
        "GeminiLLMClient": "dbgpt.model.proxy.llms.gemini",
        "AimlapiLLMClient": "dbgpt.model.proxy.llms.aimlapi",
        "SiliconFlowLLMClient": "dbgpt.model.proxy.llms.siliconflow",
        "SparkLLMClient": "dbgpt.model.proxy.llms.spark",
        "TongyiLLMClient": "dbgpt.model.proxy.llms.tongyi",
        "WenxinLLMClient": "dbgpt.model.proxy.llms.wenxin",
        "ZhipuLLMClient": "dbgpt.model.proxy.llms.zhipu",
        "YiLLMClient": "dbgpt.model.proxy.llms.yi",
        "MoonshotLLMClient": "dbgpt.model.proxy.llms.moonshot",
        "OllamaLLMClient": "dbgpt.model.proxy.llms.ollama",
        "DeepseekLLMClient": "dbgpt.model.proxy.llms.deepseek",
        "GiteeLLMClient": "dbgpt.model.proxy.llms.gitee",
        "InfiniAILLMClient": "dbgpt.model.proxy.llms.infiniai",
        "MiniMaxLLMClient": "dbgpt.model.proxy.llms.minimax",
    }

    if name in module_path:
        module = __import__(module_path[name], fromlist=[name])
        return getattr(module, name)
    else:
        raise AttributeError(f"module {__name__} has no attribute {name}")


def __getattr__(name):
    return __lazy_import(name)


__all__ = [
    "OpenAILLMClient",
    "BurnCloudLLMClient",
    "ClaudeLLMClient",
    "GeminiLLMClient",
    "TongyiLLMClient",
    "ZhipuLLMClient",
    "WenxinLLMClient",
    "AimlapiLLMClient",
    "SiliconFlowLLMClient",
    "SparkLLMClient",
    "YiLLMClient",
    "MoonshotLLMClient",
    "OllamaLLMClient",
    "DeepseekLLMClient",
    "GiteeLLMClient",
    "InfiniAILLMClient",
    "MiniMaxLLMClient",
]



"""
There are several limitations to privatizing large models: high deployment costs
and poor performance.
In scenarios where data data_privacy requirements are relatively low, connecting
with commercial large models can enable
rapid and efficient product implementation with high quality.
"""



import argparse
import os
from collections import OrderedDict
from dataclasses import MISSING, asdict, dataclass, field, fields, is_dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from dbgpt.util.annotations import Deprecated

if TYPE_CHECKING:
    from dbgpt._private.pydantic import BaseModel

T = TypeVar("T")

MISSING_DEFAULT_VALUE = "__MISSING_DEFAULT_VALUE__"

_DEFAULT_PRIVACY_FIELDS = {
    "password",
    "token",
    "key",
    "secret",
    "credential",
    "api_key",
    "api_secret",
    "db_password",
}


@dataclass
class ParameterDescription:
    """Description of a parameter in a dataclass.

    It supports nested fields for dataclass types. In some cases, we have a database
    configuration, and it has two implementations, one for MySQL and another for
    PostgreSQL, so the nested_fields will include the type of the database and its
    fields, so the key will be the type of the database and the value will be the
    fields of the database.

    Attributes:
        required: Whether this parameter is required
        is_array: Whether this parameter is an array type
        param_name: Name of the parameter
        param_class: Full path of the parameter's class (module.class_name)
        param_type: Type of the parameter
        label: Display label for the parameter
        description: Description of the parameter
        default_value: Default value of the parameter
        valid_values: List of valid values for this parameter
        ext_metadata: Additional metadata for the parameter
        nested_fields: Nested fields if this parameter is a dataclass.
    """

    required: bool = False
    is_array: bool = False
    param_name: Optional[str] = None
    param_class: Optional[str] = None
    param_type: Optional[str] = None
    label: Optional[str] = None
    description: Optional[str] = None
    default_value: Optional[Any] = None
    valid_values: Optional[List[Any]] = None
    ext_metadata: Optional[Dict[str, Any]] = None
    nested_fields: Optional[Dict[str, List["ParameterDescription"]]] = None
    param_order: Optional[int] = None

    def to_dict(self) -> Dict[str, Any]:
        """Convert the parameter description to a dictionary."""
        return asdict(self)


@dataclass
class BaseParameters:
    @classmethod
    def from_dict(cls: Type[T], data: dict, ignore_extra_fields: bool = False) -> T:
        """Create an instance of the dataclass from a dictionary.

        Args:
            data: A dictionary containing values for the dataclass fields.
            ignore_extra_fields: If True, any extra fields in the data dictionary that
                are not part of the dataclass will be ignored.
                If False, extra fields will raise an error. Defaults to False.
        Returns:
            An instance of the dataclass with values populated from the given
                dictionary.

        Raises:
            TypeError: If `ignore_extra_fields` is False and there are fields in the
                           dictionary that aren't present in the dataclass.
        """
        all_field_names = {f.name for f in fields(cls)}
        if ignore_extra_fields:
            data = {
                key: value
                for key, value in data.items()
                if key in all_field_names and value is not None
            }
        else:
            extra_fields = set(data.keys()) - all_field_names
            if extra_fields:
                raise TypeError(f"Unexpected fields: {', '.join(extra_fields)}")
        return cls(**data)

    def update_from(self, source: Union["BaseParameters", dict]) -> bool:
        """
        Update the attributes of this object using the values from another object
        (of the same or parent type) or a dictionary.
        Only update if the new value is different from the current value and the field
        is not marked as "fixed" in metadata.

        Args:
            source (Union[BaseParameters, dict]): The source to update from. Can be
                another object of the same type or a dictionary.

        Returns:
            bool: True if at least one field was updated, otherwise False.
        """
        updated = False  # Flag to indicate whether any field was updated
        if isinstance(source, (BaseParameters, dict)):
            for field_info in fields(self):
                # Check if the field has a "fixed" tag in metadata
                tags = field_info.metadata.get("tags")
                tags = [] if not tags else tags.split(",")
                if tags and "fixed" in tags:
                    continue  # skip this field
                # Get the new value from source (either another BaseParameters object
                # or a dict)
                new_value = (
                    getattr(source, field_info.name)
                    if isinstance(source, BaseParameters)
                    else source.get(field_info.name, None)
                )

                # If the new value is not None and different from the current value,
                # update the field and set the flag
                if new_value is not None and new_value != getattr(
                    self, field_info.name
                ):
                    setattr(self, field_info.name, new_value)
                    updated = True
        else:
            raise ValueError(
                "Source must be an instance of BaseParameters (or its derived class) "
                "or a dictionary."
            )

        return updated

    def __str__(self) -> str:
        return _get_dataclass_print_str(self)

    def to_command_args(self, args_prefix: str = "--") -> List[str]:
        """Convert the fields of the dataclass to a list of command line arguments.

        Args:
            args_prefix: args prefix
        Returns:
            A list of strings where each field is represented by two items:
            one for the field name prefixed by args_prefix, and one for its value.
        """
        return _dict_to_command_args(asdict(self), args_prefix=args_prefix)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @classmethod
    def get_parameter_descriptions(cls) -> List[ParameterDescription]:
        """Get the descriptions of the parameters in the dataclass."""
        return _get_parameter_descriptions(cls)

    @classmethod
    def _resolve_root_path(
        cls, path: Optional[str] = None, root_path: Optional[str] = None
    ) -> Optional[str]:
        """Replace the path to absolute path."""
        from dbgpt.configs.model_config import resolve_root_path

        return resolve_root_path(path, root_path)


def _get_dataclass_print_str(obj):
    class_name = obj.__class__.__name__
    parameters = [
        f"\n\n=========================== {class_name} ===========================\n"
    ]
    for field_info in fields(obj):
        value = _get_simple_privacy_field_value(obj, field_info)
        parameters.append(f"{field_info.name}: {value}")
    parameters.append(
        "\n======================================================================\n\n"
    )
    return "\n".join(parameters)


def _dict_to_command_args(obj: Dict, args_prefix: str = "--") -> List[str]:
    """Convert dict to a list of command line arguments

    Args:
        obj: dict
    Returns:
        A list of strings where each field is represented by two items:
        one for the field name prefixed by args_prefix, and one for its value.
    """
    args = []
    for key, value in obj.items():
        if value is None:
            continue
        args.append(f"{args_prefix}{key}")
        args.append(str(value))
    return args


def _get_simple_privacy_field_value(
    obj, field_info, sensitive_fields: Optional[set] = None
):
    """Retrieve the value of a field from a dataclass instance, applying privacy rules
    if necessary.

    This function reads the metadata of a field to check if it's tagged with 'privacy'.
    If the 'privacy' tag is present, then it modifies the value based on its type
    for privacy concerns:
    - int: returns -999
    - float: returns -999.0
    - bool: returns False
    - str: if length > 5, masks the middle part and returns first and last char;
           otherwise, returns "******"

    Args:
        obj: The dataclass instance.
        field_info: A Field object that contains information about the dataclass field.

    Returns:
    The original or modified value of the field based on the privacy rules.

    Example usage:
    @dataclass
    class Person:
        name: str
        age: int
        ssn: str = field(metadata={"tags": "privacy"})
    p = Person("Alice", 30, "123-45-6789")
    print(_get_simple_privacy_field_value(p, Person.ssn))  # A******9
    """
    if sensitive_fields is None:
        sensitive_fields = _DEFAULT_PRIVACY_FIELDS
    tags = field_info.metadata.get("tags")
    tags = [] if not tags else tags.split(",")
    is_privacy = False
    if tags and "privacy" in tags:
        is_privacy = True
    elif field_info.name in sensitive_fields:
        is_privacy = True
    value = getattr(obj, field_info.name)
    if not is_privacy or not value:
        return value
    field_type = EnvArgumentParser._get_argparse_type(field_info.type)
    if field_type is int:
        return -999
    if field_type is float:
        return -999.0
    if field_type is bool:
        return False
    # str
    if len(value) > 5:
        return "".join(value[:2]) + "******" + "".join(value[-2:])
    return "******"


def _genenv_ignoring_key_case(
    env_key: str, env_prefix: Optional[str] = None, default_value: Optional[str] = None
):
    """Get the value from the environment variable, ignoring the case of the key"""
    if env_prefix:
        env_key = env_prefix + env_key
    return os.getenv(
        env_key, os.getenv(env_key.upper(), os.getenv(env_key.lower(), default_value))
    )


def _genenv_ignoring_key_case_with_prefixes(
    env_key: str,
    env_prefixes: Optional[List[str]] = None,
    default_value: Optional[str] = None,
) -> str:
    if env_prefixes:
        for env_prefix in env_prefixes:
            env_var_value = _genenv_ignoring_key_case(env_key, env_prefix)
            if env_var_value:
                return env_var_value
    return _genenv_ignoring_key_case(env_key, default_value=default_value)


class EnvArgumentParser:
    @staticmethod
    def get_env_prefix(env_key: str) -> Optional[str]:
        if not env_key:
            return None
        env_key = env_key.replace("-", "_")
        return env_key + "_"

    def parse_args_into_dataclass(
        self,
        dataclass_type: Type,
        env_prefixes: Optional[List[str]] = None,
        command_args: Optional[List[str]] = None,
        **kwargs,
    ) -> Any:
        """Parse parameters from environment variables and command lines and populate
        them into data class.
        """
        parser = argparse.ArgumentParser(allow_abbrev=False)
        for fd in fields(dataclass_type):
            env_var_value: Any = _genenv_ignoring_key_case_with_prefixes(
                fd.name, env_prefixes
            )
            if env_var_value:
                env_var_value = env_var_value.strip()
                if fd.type is int or fd.type == Optional[int]:
                    env_var_value = int(env_var_value)
                elif fd.type is float or fd.type == Optional[float]:
                    env_var_value = float(env_var_value)
                elif fd.type is bool or fd.type == Optional[bool]:
                    env_var_value = env_var_value.lower() == "true"
                elif fd.type is str or fd.type == Optional[str]:
                    pass
                else:
                    raise ValueError(f"Unsupported parameter type {fd.type}")
            if not env_var_value:
                env_var_value = kwargs.get(fd.name)

            # print(f"env_var_value: {env_var_value} for {field.name}")
            # Add a command-line argument for this field
            EnvArgumentParser._build_single_argparse_option(parser, fd, env_var_value)

        # Parse the command-line arguments
        cmd_args, cmd_argv = parser.parse_known_args(args=command_args)
        # cmd_args = parser.parse_args(args=command_args)
        # print(f"cmd_args: {cmd_args}")
        for fd in fields(dataclass_type):
            # cmd_line_value = getattr(cmd_args, field.name)
            if fd.name in cmd_args:
                cmd_line_value = getattr(cmd_args, fd.name)
                if cmd_line_value is not None:
                    kwargs[fd.name] = cmd_line_value

        return dataclass_type(**kwargs)

    @staticmethod
    def _create_arg_parser(dataclass_type: Type) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(description=dataclass_type.__doc__)
        for fd in fields(dataclass_type):
            help_text = fd.metadata.get("help", "")
            valid_values = fd.metadata.get("valid_values", None)
            argument_kwargs = {
                "type": EnvArgumentParser._get_argparse_type(fd.type),
                "help": help_text,
                "choices": valid_values,
                "required": EnvArgumentParser._is_require_type(fd.type),
            }
            if fd.default != MISSING:
                argument_kwargs["default"] = fd.default
                argument_kwargs["required"] = False
            parser.add_argument(f"--{fd.name}", **argument_kwargs)
        return parser

    @staticmethod
    def _create_click_option_from_field(field_name: str, fd: Type, is_func=True):
        import click

        help_text = fd.metadata.get("help", "")
        valid_values = fd.metadata.get("valid_values", None)
        cli_params = {
            "default": None if fd.default is MISSING else fd.default,
            "help": help_text,
            "show_default": True,
            "required": fd.default is MISSING,
        }
        if valid_values:
            cli_params["type"] = click.Choice(valid_values)
        real_type = EnvArgumentParser._get_argparse_type(fd.type)
        if real_type is int:
            cli_params["type"] = click.INT
        elif real_type is float:
            cli_params["type"] = click.FLOAT
        elif real_type is str:
            cli_params["type"] = click.STRING
        elif real_type is bool:
            cli_params["is_flag"] = True
        name = f"--{field_name}"
        if is_func:
            return click.option(
                name,
                **cli_params,
            )
        else:
            return click.Option([name], **cli_params)

    @staticmethod
    def create_click_option(
        *dataclass_types: Type,
        _dynamic_factory: Optional[Callable[[], List[Type]]] = None,
    ):
        import functools
        from collections import OrderedDict

        combined_fields = OrderedDict()
        if _dynamic_factory:
            _types = _dynamic_factory()
            if _types:
                dataclass_types = list(_types)  # type: ignore
        for dataclass_type in dataclass_types:
            # type: ignore
            for fd in fields(dataclass_type):
                if fd.name not in combined_fields:
                    combined_fields[fd.name] = fd

        def decorator(func):
            for field_name, fd in reversed(combined_fields.items()):
                option_decorator = EnvArgumentParser._create_click_option_from_field(
                    field_name, fd
                )
                func = option_decorator(func)

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    @staticmethod
    def _create_raw_click_option(
        *dataclass_types: Type,
        _dynamic_factory: Optional[Callable[[], List[Type]]] = None,
    ):
        combined_fields = _merge_dataclass_types(
            *dataclass_types, _dynamic_factory=_dynamic_factory
        )
        options = []

        for field_name, fd in reversed(combined_fields.items()):
            options.append(
                EnvArgumentParser._create_click_option_from_field(
                    field_name, fd, is_func=False
                )
            )
        return options

    @staticmethod
    def create_argparse_option(
        *dataclass_types: Type,
        _dynamic_factory: Optional[Callable[[], List[Type]]] = None,
    ) -> argparse.ArgumentParser:
        combined_fields = _merge_dataclass_types(
            *dataclass_types, _dynamic_factory=_dynamic_factory
        )
        parser = argparse.ArgumentParser()
        for _, fd in reversed(combined_fields.items()):
            EnvArgumentParser._build_single_argparse_option(parser, fd)
        return parser

    @staticmethod
    def _build_single_argparse_option(
        parser: argparse.ArgumentParser, field, default_value=None
    ):
        # Add a command-line argument for this field
        help_text = field.metadata.get("help", "")
        valid_values = field.metadata.get("valid_values", None)
        short_name = field.metadata.get("short", None)
        argument_kwargs = {
            "type": EnvArgumentParser._get_argparse_type(field.type),
            "help": help_text,
            "choices": valid_values,
            "required": EnvArgumentParser._is_require_type(field.type),
        }
        if field.default != MISSING:
            argument_kwargs["default"] = field.default
            argument_kwargs["required"] = False
        if default_value:
            argument_kwargs["default"] = default_value
            argument_kwargs["required"] = False
        if field.type is bool or field.type == Optional[bool]:
            argument_kwargs["action"] = "store_true"
            del argument_kwargs["type"]
            del argument_kwargs["choices"]
        names = []
        if short_name:
            names.append(f"-{short_name}")
        names.append(f"--{field.name}")
        parser.add_argument(*names, **argument_kwargs)

    @staticmethod
    def _get_argparse_type(
        field_type: Type, only_support_base_type: bool = True
    ) -> Type:
        # Return the appropriate type for argparse to use based on the field type
        if field_type is int or field_type == Optional[int]:
            return int
        elif field_type is float or field_type == Optional[float]:
            return float
        elif field_type is bool or field_type == Optional[bool]:
            return bool
        elif field_type is str or field_type == Optional[str]:
            return str
        elif field_type is dict or field_type == Optional[dict]:
            return dict
        elif only_support_base_type:
            raise ValueError(f"Unsupported parameter type {field_type}")
        else:
            return field_type

    @staticmethod
    def _get_argparse_type_str(field_type: Type, only_support_base_type: bool) -> str:
        from .function_utils import type_to_string

        argparse_type = EnvArgumentParser._get_argparse_type(
            field_type, only_support_base_type
        )
        str_type, sub_types = type_to_string(field_type)

        if argparse_type is int:
            return "int"
        elif argparse_type is float:
            return "float"
        elif argparse_type is bool:
            return "bool"
        elif argparse_type is dict:
            return "dict"
        elif argparse_type is str:
            return "str"
        elif only_support_base_type:
            raise ValueError(f"Unsupported parameter type {field_type}")
        else:
            str_type, sub_types = type_to_string(field_type)
            return field_type.__name__

    @staticmethod
    def _is_require_type(field_type: Type) -> bool:
        return field_type not in [Optional[int], Optional[float], Optional[bool]]

    @staticmethod
    def _kwargs_to_env_key_value(
        kwargs: Dict, prefix: str = "__dbgpt_gunicorn__env_prefix__"
    ) -> Dict[str, str]:
        return {prefix + k: str(v) for k, v in kwargs.items()}

    @staticmethod
    def _read_env_key_value(
        prefix: str = "__dbgpt_gunicorn__env_prefix__",
    ) -> List[str]:
        env_args = []
        for key, value in os.environ.items():
            if key.startswith(prefix):
                arg_key = "--" + key.replace(prefix, "")
                if value.lower() in ["true", "1"]:
                    # Flag args
                    env_args.append(arg_key)
                elif value.lower() not in ["false", "0"]:
                    env_args.extend([arg_key, value])
        return env_args


def _merge_dataclass_types(
    *dataclass_types: Type, _dynamic_factory: Optional[Callable[[], List[Type]]] = None
) -> OrderedDict:
    combined_fields = OrderedDict()
    if _dynamic_factory:
        _types = _dynamic_factory()
        if _types:
            dataclass_types = list(_types)  # type: ignore
    for dataclass_type in dataclass_types:
        for fd in fields(dataclass_type):
            if fd.name not in combined_fields:
                combined_fields[fd.name] = fd
    return combined_fields


def _type_str_to_python_type(type_str: str) -> Type:
    type_mapping: Dict[str, Type] = {
        "int": int,
        "float": float,
        "bool": bool,
        "str": str,
    }
    return type_mapping.get(type_str, str)


def _get_parameter_descriptions(
    dataclass_type: Type, parent_field: Optional[str] = None, **kwargs
) -> List[ParameterDescription]:
    """Get the descriptions of the parameters in the dataclass with nested field
    support.

    Args:
        dataclass_type: The dataclass type to get descriptions for
        parent_field: Name of the parent field if this is a nested parameter
        **kwargs: Additional keyword arguments

    Returns:
        List of ParameterDescription objects describing all fields including nested ones
    """

    from dbgpt.util.configure.manager import ConfigurationManager

    return ConfigurationManager.parse_description(dataclass_type)


def _build_parameter_class(desc: List[ParameterDescription]) -> Type:
    from dbgpt.util.module_utils import import_from_string

    if not desc:
        raise ValueError("Parameter descriptions cant be empty")
    param_class_str = desc[0].param_class
    class_name = None
    if param_class_str:
        param_class = import_from_string(param_class_str, ignore_import_error=True)
        if param_class:
            return param_class
        module_name, _, class_name = param_class_str.rpartition(".")

    fields_dict = {}  # This will store field names and their default values or field()
    annotations = {}  # This will store the type annotations for the fields

    for d in desc:
        metadata = d.ext_metadata if d.ext_metadata else {}
        metadata["help"] = d.description
        metadata["valid_values"] = d.valid_values

        annotations[d.param_name] = _type_str_to_python_type(
            d.param_type  # type: ignore
        )  # Set type annotation
        # fields_dict[d.param_name] = field(default=d.default_value, metadata=metadata)
        if d.param_name == "ignore_patterns":
            fields_dict[d.param_name] = field(default=None, metadata=metadata)
        else:
            fields_dict[d.param_name] = field(
                default=d.default_value, metadata=metadata
            )

    # Create the new class. Note the setting of __annotations__ for type hints
    new_class = type(
        class_name,  # type: ignore
        (object,),
        {**fields_dict, "__annotations__": annotations},  # type: ignore
    )
    # Make it a dataclass
    result_class = dataclass(new_class)  # type: ignore

    return result_class


@Deprecated(version="0.7.0", remove_version="0.8.0")
def _extract_parameter_details(
    parser: argparse.ArgumentParser,
    param_class: Optional[str] = None,
    skip_names: Optional[List[str]] = None,
    overwrite_default_values: Optional[Dict[str, Any]] = None,
) -> List[ParameterDescription]:
    from .function_utils import format_type_string

    if overwrite_default_values is None:
        overwrite_default_values = {}
    descriptions = []

    for action in parser._actions:
        if (
            action.default == argparse.SUPPRESS
        ):  # typically this means the argument was not provided
            continue

        # determine parameter class (store_true/store_false are flags)
        # flag_or_option = (
        #     "flag" if isinstance(action, argparse._StoreConstAction) else "option"
        # )

        # extract parameter name (use the first option string, typically the long form)
        param_name = action.option_strings[0] if action.option_strings else action.dest
        if param_name.startswith("--"):
            param_name = param_name[2:]
        if param_name.startswith("-"):
            param_name = param_name[1:]

        param_name = param_name.replace("-", "_")

        if skip_names and param_name in skip_names:
            continue

        # gather other details
        default_value = action.default
        if param_name in overwrite_default_values:
            default_value = overwrite_default_values[param_name]
        arg_type = (
            action.type if not callable(action.type) else str(action.type.__name__)  # type: ignore
        )
        arg_type = format_type_string(arg_type)
        description = action.help

        # determine if the argument is required
        required = action.required

        # extract valid values for choices, if provided
        valid_values = list(action.choices) if action.choices is not None else None

        # set ext_metadata as an empty dict for now, can be updated later if needed
        ext_metadata: Dict[str, Any] = {}

        descriptions.append(
            ParameterDescription(
                param_class=param_class,
                param_name=param_name,
                param_type=arg_type,
                default_value=default_value,
                description=description,
                required=required,
                valid_values=valid_values,
                ext_metadata=ext_metadata,
            )
        )

    return descriptions


def _get_dict_from_obj(obj, default_value=None) -> Optional[Dict]:
    if not obj:
        return None
    if is_dataclass(type(obj)):
        params = {}
        for field_info in fields(obj):
            value = _get_simple_privacy_field_value(obj, field_info)
            params[field_info.name] = value
        return params
    if isinstance(obj, dict):
        return obj
    return default_value


@Deprecated(version="0.7.0", remove_version="0.8.0")
def _get_base_model_descriptions(model_cls: "BaseModel") -> List[ParameterDescription]:
    from dbgpt._private import pydantic

    version = int(pydantic.VERSION.split(".")[0])  # type: ignore
    schema = model_cls.model_json_schema() if version >= 2 else model_cls.schema()
    required_fields = set(schema.get("required", []))
    param_descs = []
    for field_name, field_schema in schema.get("properties", {}).items():
        field = model_cls.model_fields[field_name]
        param_type = field_schema.get("type")
        if not param_type and "anyOf" in field_schema:
            for any_of in field_schema["anyOf"]:
                if any_of["type"] != "null":
                    param_type = any_of["type"]
                    break
        if version >= 2:
            default_value = (
                field.default
                if hasattr(field, "default")
                and str(field.default) != "PydanticUndefined"
                else None
            )
        else:
            default_value = (
                field.default
                if not field.allow_none
                else (
                    field.default_factory() if callable(field.default_factory) else None
                )
            )
        description = field_schema.get("description", "")
        is_required = field_name in required_fields
        valid_values = None
        ext_metadata = None
        if hasattr(field, "field_info"):
            valid_values = (
                list(field.field_info.choices)
                if hasattr(field.field_info, "choices")
                else None
            )
            ext_metadata = (
                field.field_info.extra if hasattr(field.field_info, "extra") else None
            )
        param_class = f"{model_cls.__module__}.{model_cls.__name__}"
        param_desc = ParameterDescription(
            param_class=param_class,
            param_name=field_name,
            param_type=param_type,
            default_value=default_value,
            description=description,
            required=is_required,
            valid_values=valid_values,
            ext_metadata=ext_metadata,
        )
        param_descs.append(param_desc)
    return param_descs


class _SimpleArgParser:
    def __init__(self, *args):
        self.params = {arg.replace("_", "-"): None for arg in args}

    def parse(self, args=None):
        import sys

        if args is None:
            args = sys.argv[1:]
        else:
            args = list(args)
        prev_arg = None
        for arg in args:
            if arg.startswith("--"):
                if prev_arg:
                    self.params[prev_arg] = None
                prev_arg = arg[2:]
            else:
                if prev_arg:
                    self.params[prev_arg] = arg
                    prev_arg = None

        if prev_arg:
            self.params[prev_arg] = None

    def _get_param(self, key):
        return self.params.get(key.replace("_", "-")) or self.params.get(key)

    def __getattr__(self, item):
        return self._get_param(item)

    def __getitem__(self, key):
        return self._get_param(key)

    def get(self, key, default=None):
        return self._get_param(key) or default

    def __str__(self):
        return "\n".join(
            [f"{key.replace('-', '_')}: {value}" for key, value in self.params.items()]
        )


def build_lazy_click_command(*dataclass_types: Type, _dynamic_factory=None):
    import click

    class LazyCommand(click.Command):
        def __init__(self, *args, **kwargs):
            super(LazyCommand, self).__init__(*args, **kwargs)
            self.dynamic_params_added = False

        def get_params(self, ctx):
            if ctx and not self.dynamic_params_added:
                dynamic_params = EnvArgumentParser._create_raw_click_option(
                    *dataclass_types, _dynamic_factory=_dynamic_factory
                )
                for param in reversed(dynamic_params):
                    self.params.append(param)
                self.dynamic_params_added = True
            return super(LazyCommand, self).get_params(ctx)

    return LazyCommand



import logging
import time
from dataclasses import dataclass, field
from typing import Dict, List, Optional

logger = logging.getLogger(__name__)


@dataclass
class LLMPerformanceMetrics:
    """Performance metrics for LLM inference, including prefill and decode phases"""

    # Token counts
    input_token_count: int = 0
    total_tokens_generated: int = 0
    prev_tokens_count: int = 0

    # Time measurements in nanoseconds
    start_time_ns: int = field(default_factory=time.time_ns)
    prefill_start_time_ns: Optional[int] = None
    prefill_end_time_ns: Optional[int] = None
    prefill_time_ns: Optional[int] = None
    decode_start_time_ns: Optional[int] = None
    total_time_ns: Optional[int] = None

    # Timestamps and measurements
    token_timestamps_ns: List[int] = field(default_factory=list)
    decode_times_ns: List[int] = field(default_factory=list)

    # Calculated metrics (tokens per second)
    prefill_tokens_per_second: Optional[float] = None
    decode_tokens_per_second: Optional[float] = None
    end_to_end_tokens_per_second: Optional[float] = None

    # Additional computed values
    avg_decode_time: Optional[float] = None

    def to_dict(self) -> Dict[str, any]:
        """Convert metrics to a dictionary for API response, with times in seconds"""
        metrics = {
            "input_token_count": self.input_token_count,
            "total_tokens_generated": self.total_tokens_generated,
        }

        # Add time metrics in seconds
        if self.prefill_time_ns is not None:
            metrics["prefill_time"] = self.prefill_time_ns / 1e9
            metrics["prefill_tokens_per_second"] = self.prefill_tokens_per_second or 0

        if self.total_time_ns is not None:
            metrics["total_time"] = self.total_time_ns / 1e9

        if self.decode_times_ns:
            metrics["avg_decode_time"] = self.avg_decode_time
            metrics["decode_tokens_per_second"] = self.decode_tokens_per_second or 0

        # Add throughput metrics
        metrics["end_to_end_tokens_per_second"] = self.end_to_end_tokens_per_second or 0

        return metrics


class LLMPerformanceMonitor:
    """Generic performance monitor for LLM inference that tracks prefill and decode
    phases"""

    def __init__(self, input_token_count: int = 0):
        # Performance metrics
        self.metrics = LLMPerformanceMetrics(input_token_count=input_token_count)

        # Phase flags
        self.prefill_started: bool = False
        self.first_token_received: bool = False

    def start_prefill(self) -> int:
        """Mark the beginning of the prefill phase using nanosecond timestamp"""
        timestamp = time.time_ns()
        self.metrics.prefill_start_time_ns = timestamp
        self.prefill_started = True
        return timestamp

    def on_tokens_received(self, current_token_count: int) -> Dict[str, any]:
        """
        Called when new tokens are received from LLM
        Returns updated performance metrics
        """
        current_time_ns = time.time_ns()

        # Calculate new tokens received in this batch
        new_tokens = current_token_count - self.metrics.prev_tokens_count

        # Auto-detect the end of prefill / start of decode phase
        if self.prefill_started and not self.first_token_received and new_tokens > 0:
            # This is the first tokens batch - mark the end of prefill phase
            self.metrics.prefill_end_time_ns = current_time_ns
            self.metrics.prefill_time_ns = (
                self.metrics.prefill_end_time_ns - self.metrics.prefill_start_time_ns
            )

            # Convert nanoseconds to seconds for calculation and logging
            prefill_time_sec = self.metrics.prefill_time_ns / 1e9

            # Calculate prefill speed
            if self.metrics.input_token_count > 0 and prefill_time_sec > 0:
                self.metrics.prefill_tokens_per_second = (
                    self.metrics.input_token_count / prefill_time_sec
                )
                logger.info(
                    f"Prefill speed: {self.metrics.prefill_tokens_per_second:.2f} "
                    f"tokens/s for {self.metrics.input_token_count} tokens"
                )

            # Mark the beginning of decode phase
            self.metrics.decode_start_time_ns = current_time_ns
            self.first_token_received = True

        # Record token generation data
        if self.first_token_received and new_tokens > 0:
            # If we've already received tokens, add decode time for this batch
            if len(self.metrics.token_timestamps_ns) > 0:
                last_timestamp_ns = self.metrics.token_timestamps_ns[-1]
                token_decode_time_ns = current_time_ns - last_timestamp_ns

                # Distribute the time evenly across all new tokens in this batch
                time_per_token = token_decode_time_ns / new_tokens
                for _ in range(new_tokens):
                    self.metrics.decode_times_ns.append(time_per_token)

        # Record the current token batch timestamp
        self.metrics.token_timestamps_ns.append(current_time_ns)
        self.metrics.total_tokens_generated += new_tokens
        self.metrics.prev_tokens_count = current_token_count

        # Calculate current metrics
        self._update_metrics(current_time_ns)

        return self.get_metrics_dict()

    def _update_metrics(self, current_time_ns: Optional[int] = None) -> None:
        """Update the performance metrics based on current state"""
        if current_time_ns is None:
            current_time_ns = time.time_ns()

        # Record total time
        self.metrics.total_time_ns = current_time_ns - self.metrics.start_time_ns

        # Calculate average decode speed
        if self.metrics.decode_times_ns:
            # Convert to seconds
            decode_times_sec = [t / 1e9 for t in self.metrics.decode_times_ns]
            self.metrics.avg_decode_time = sum(decode_times_sec) / len(decode_times_sec)
            self.metrics.decode_tokens_per_second = 1.0 / self.metrics.avg_decode_time

        # Calculate end-to-end throughput
        total_time_sec = self.metrics.total_time_ns / 1e9
        if total_time_sec > 0:
            total_tokens = (
                self.metrics.input_token_count + self.metrics.total_tokens_generated
            )
            self.metrics.end_to_end_tokens_per_second = total_tokens / total_time_sec

    def end_generation(self) -> Dict[str, any]:
        """Mark the end of generation and finalize metrics"""
        current_time_ns = time.time_ns()
        self._update_metrics(current_time_ns)

        # Log final performance data
        total_time_sec = self.metrics.total_time_ns / 1e9
        logger.info(f"Generation complete. Total time: {total_time_sec:.6f}s")

        if self.metrics.prefill_tokens_per_second:
            logger.info(
                f"Final prefill speed: {self.metrics.prefill_tokens_per_second:.2f} "
                "tokens/s"
            )

        if self.metrics.decode_tokens_per_second:
            logger.info(
                f"Final decode speed: {self.metrics.decode_tokens_per_second:.2f} "
                "tokens/s"
            )

        if self.metrics.end_to_end_tokens_per_second:
            logger.info(
                "End-to-end throughput: "
                f"{self.metrics.end_to_end_tokens_per_second:.2f} tokens/s"
            )

        return self.get_metrics_dict()

    def get_metrics_dict(self) -> Dict[str, any]:
        """Get performance metrics as dictionary, converting nanoseconds to seconds
        for external use"""
        return self.metrics.to_dict()



"""Smoke tests for CLI test infrastructure — verifies fixture isolation."""

import os
from pathlib import Path


def test_isolated_dbgpt_home_is_not_real_home(isolated_dbgpt_home):
    """Verify that isolated_dbgpt_home does NOT point to the real ~/.dbgpt."""
    real_home = Path.home() / ".dbgpt"
    assert isolated_dbgpt_home != real_home, (
        f"isolated_dbgpt_home should not be the real home: {real_home}"
    )


def test_isolated_dbgpt_home_env_var(isolated_dbgpt_home):
    """Verify DBGPT_HOME env var is set to the isolated path."""
    env_home = os.environ.get("DBGPT_HOME")
    assert env_home is not None, "DBGPT_HOME env var should be set"
    assert env_home == str(isolated_dbgpt_home), (
        f"DBGPT_HOME ({env_home}) should match isolated_dbgpt_home ({isolated_dbgpt_home})"  # noqa: E501
    )


def test_isolated_dbgpt_home_patches_config_module(isolated_dbgpt_home):
    """Verify _config module constants are patched to isolated temp path."""
    import dbgpt.cli._config as _cfg

    assert _cfg._DBGPT_HOME == isolated_dbgpt_home, (
        f"_config._DBGPT_HOME ({_cfg._DBGPT_HOME}) should equal isolated home"
    )
    assert _cfg._CONFIGS_DIR == isolated_dbgpt_home / "configs", (
        "_config._CONFIGS_DIR should be under isolated home"
    )
    assert _cfg._ACTIVE_CONFIG == isolated_dbgpt_home / "config.toml", (
        "_config._ACTIVE_CONFIG should be under isolated home"
    )


def test_cli_runner_returns_click_runner(cli_runner):
    """Verify cli_runner fixture returns a CliRunner instance."""
    import click.testing

    assert isinstance(cli_runner, click.testing.CliRunner)



"""Agentic IntentTranslator class."""

import logging

from dbgpt.rag.transformer.base import TranslatorBase

logger = logging.getLogger(__name__)


class AgenticIntentTranslator(TranslatorBase):
    """Agentic IntentTranslator class."""



from dbgpt.util.pagination_utils import PaginationResult


def test_build_from_all_normal_case():
    items = list(range(100))
    result = PaginationResult.build_from_all(items, page=2, page_size=20)

    assert len(result.items) == 20
    assert result.items == list(range(20, 40))
    assert result.total_count == 100
    assert result.total_pages == 5
    assert result.page == 2
    assert result.page_size == 20


def test_build_from_all_empty_list():
    items = []
    result = PaginationResult.build_from_all(items, page=1, page_size=5)

    assert result.items == []
    assert result.total_count == 0
    assert result.total_pages == 0
    assert result.page == 0
    assert result.page_size == 5


def test_build_from_all_last_page():
    items = list(range(95))
    result = PaginationResult.build_from_all(items, page=5, page_size=20)

    assert len(result.items) == 15
    assert result.items == list(range(80, 95))
    assert result.total_count == 95
    assert result.total_pages == 5
    assert result.page == 5
    assert result.page_size == 20


def test_build_from_all_page_out_of_range():
    items = list(range(50))
    result = PaginationResult.build_from_all(items, page=10, page_size=10)

    assert len(result.items) == 10
    assert result.items == list(range(40, 50))
    assert result.total_count == 50
    assert result.total_pages == 5
    assert result.page == 5
    assert result.page_size == 10


def test_build_from_all_page_zero():
    items = list(range(50))
    result = PaginationResult.build_from_all(items, page=0, page_size=10)

    assert len(result.items) == 10
    assert result.items == list(range(0, 10))
    assert result.total_count == 50
    assert result.total_pages == 5
    assert result.page == 1
    assert result.page_size == 10


def test_build_from_all_negative_page():
    items = list(range(50))
    result = PaginationResult.build_from_all(items, page=-1, page_size=10)

    assert len(result.items) == 10
    assert result.items == list(range(0, 10))
    assert result.total_count == 50
    assert result.total_pages == 5
    assert result.page == 1
    assert result.page_size == 10


def test_build_from_all_page_size_larger_than_total():
    items = list(range(50))
    result = PaginationResult.build_from_all(items, page=1, page_size=100)

    assert len(result.items) == 50
    assert result.items == list(range(50))
    assert result.total_count == 50
    assert result.total_pages == 1
    assert result.page == 1
    assert result.page_size == 100



"""Protocol for the dashboard vis."""

import json
import logging
from typing import Any, Dict, Optional

from ..base import Vis

logger = logging.getLogger(__name__)


class VisDashboard(Vis):
    """Dashboard Vis Protocol."""

    def sync_generate_param(self, **kwargs) -> Optional[Dict[str, Any]]:
        """Generate the parameters required by the vis protocol."""
        charts: Optional[dict] = kwargs.get("charts", None)
        title: Optional[str] = kwargs.get("title", None)
        if not charts:
            raise ValueError(
                f"Parameter information is missing and {self.vis_tag} protocol "
                "conversion cannot be performed."
            )

        chart_items = []
        for chart in charts:
            param = {}
            sql = chart.get("sql", "")
            param["sql"] = sql
            param["type"] = chart.get("display_type", "response_table")
            param["title"] = chart.get("title", "")
            param["describe"] = chart.get("thought", "")
            try:
                df = chart.get("data", None)
                err_msg = chart.get("err_msg", None)
                if df is None:
                    param["err_msg"] = err_msg
                else:
                    param["data"] = json.loads(
                        df.to_json(orient="records", date_format="iso", date_unit="s")
                    )
            except Exception as e:
                logger.exception("dashboard chart build faild！")
                param["data"] = []
                param["err_msg"] = str(e)
            chart_items.append(param)

        dashboard_param = {
            "data": chart_items,
            "chart_count": len(chart_items),
            "title": title,
            "display_strategy": "default",
            "style": "default",
        }

        return dashboard_param

    @classmethod
    def vis_tag(cls):
        """Vis Dashboard."""
        return "vis-dashboard"



"""File storage interface."""

import dataclasses
import hashlib
import io
import logging
import os
import uuid
from abc import ABC, abstractmethod
from io import BytesIO
from typing import Any, BinaryIO, Dict, List, Optional, Tuple
from urllib.parse import parse_qs, urlencode, urlparse

import requests

from dbgpt.component import BaseComponent, ComponentType, SystemApp
from dbgpt.util.tracer import root_tracer, trace

from ...util import BaseParameters, RegisterParameters
from .storage import (
    InMemoryStorage,
    QuerySpec,
    ResourceIdentifier,
    StorageError,
    StorageInterface,
    StorageItem,
)

logger = logging.getLogger(__name__)
_SCHEMA = "dbgpt-fs"


@dataclasses.dataclass
class FileMetadataIdentifier(ResourceIdentifier):
    """File metadata identifier."""

    file_id: str
    bucket: str

    def to_dict(self) -> Dict:
        """Convert the identifier to a dictionary."""
        return {"file_id": self.file_id, "bucket": self.bucket}

    @property
    def str_identifier(self) -> str:
        """Get the string identifier.

        Returns:
            str: The string identifier
        """
        return f"{self.bucket}/{self.file_id}"


@dataclasses.dataclass
class FileMetadata(StorageItem):
    """File metadata for storage."""

    file_id: str
    bucket: str
    file_name: str
    file_size: int
    storage_type: str
    storage_path: str
    uri: str
    custom_metadata: Dict[str, Any]
    file_hash: str
    user_name: Optional[str] = None
    sys_code: Optional[str] = None
    _identifier: FileMetadataIdentifier = dataclasses.field(init=False)

    def __post_init__(self):
        """Post init method."""
        self._identifier = FileMetadataIdentifier(
            file_id=self.file_id, bucket=self.bucket
        )
        custom_metadata = self.custom_metadata or {}
        if not self.user_name:
            self.user_name = custom_metadata.get("user_name")
        if not self.sys_code:
            self.sys_code = custom_metadata.get("sys_code")

    @property
    def identifier(self) -> ResourceIdentifier:
        """Get the resource identifier."""
        return self._identifier

    def merge(self, other: "StorageItem") -> None:
        """Merge the metadata with another item."""
        if not isinstance(other, FileMetadata):
            raise StorageError("Cannot merge different types of items")
        self._from_object(other)

    def to_dict(self) -> Dict:
        """Convert the metadata to a dictionary."""
        return {
            "file_id": self.file_id,
            "bucket": self.bucket,
            "file_name": self.file_name,
            "file_size": self.file_size,
            "storage_type": self.storage_type,
            "storage_path": self.storage_path,
            "uri": self.uri,
            "custom_metadata": self.custom_metadata,
            "file_hash": self.file_hash,
        }

    def _from_object(self, obj: "FileMetadata") -> None:
        self.file_id = obj.file_id
        self.bucket = obj.bucket
        self.file_name = obj.file_name
        self.file_size = obj.file_size
        self.storage_type = obj.storage_type
        self.storage_path = obj.storage_path
        self.uri = obj.uri
        self.custom_metadata = obj.custom_metadata
        self.file_hash = obj.file_hash
        self._identifier = obj._identifier


@dataclasses.dataclass
class StorageBackendConfig(BaseParameters, RegisterParameters):
    """Storage backend configuration"""

    __type__ = "___storage_backend_config___"
    __cfg_type__ = "utils"

    def create_storage(self) -> "StorageBackend":
        """Create the storage"""
        raise NotImplementedError()


class FileStorageURI:
    """File storage URI."""

    def __init__(
        self,
        storage_type: str,
        bucket: str,
        file_id: str,
        version: Optional[str] = None,
        custom_params: Optional[Dict[str, Any]] = None,
    ):
        """Initialize the file storage URI."""
        self.scheme = _SCHEMA
        self.storage_type = storage_type
        self.bucket = bucket
        self.file_id = file_id
        self.version = version
        self.custom_params = custom_params or {}

    @classmethod
    def is_local_file(cls, uri: str) -> bool:
        """Check if the URI is local."""
        parsed = urlparse(uri)
        if not parsed.scheme or parsed.scheme == "file":
            return True
        return False

    @classmethod
    def parse(cls, uri: str) -> "FileStorageURI":
        """Parse the URI string."""
        parsed = urlparse(uri)
        if parsed.scheme != _SCHEMA:
            raise ValueError(f"Invalid URI scheme. Must be '{_SCHEMA}'")
        path_parts = parsed.path.strip("/").split("/")
        if len(path_parts) < 2:
            raise ValueError("Invalid URI path. Must contain bucket and file ID")
        storage_type = parsed.netloc
        bucket = path_parts[0]
        file_id = path_parts[1]
        version = path_parts[2] if len(path_parts) > 2 else None
        custom_params = parse_qs(parsed.query)
        return cls(storage_type, bucket, file_id, version, custom_params)

    def __str__(self) -> str:
        """Get the string representation of the URI."""
        base_uri = f"{self.scheme}://{self.storage_type}/{self.bucket}/{self.file_id}"
        if self.version:
            base_uri += f"/{self.version}"
        if self.custom_params:
            query_string = urlencode(self.custom_params, doseq=True)
            base_uri += f"?{query_string}"
        return base_uri


class StorageBackend(ABC):
    """Storage backend interface."""

    storage_type: str = "__base__"

    @abstractmethod
    def save(
        self,
        bucket: str,
        file_id: str,
        file_data: BinaryIO,
        public_url: bool = False,
        public_url_expire: Optional[int] = None,
    ) -> str:
        """Save the file data to the storage backend.

        Args:
            bucket (str): The bucket name
            file_id (str): The file ID
            file_data (BinaryIO): The file data
            public_url (bool): Whether to generate a public URL
            public_url_expire (Optional[int]): The expiration time for the public URL

        Returns:
            str: The storage path
        """

    @abstractmethod
    def load(self, fm: FileMetadata) -> BinaryIO:
        """Load the file data from the storage backend.

        Args:
            fm (FileMetadata): The file metadata

        Returns:
            BinaryIO: The file data
        """

    @abstractmethod
    def delete(self, fm: FileMetadata) -> bool:
        """Delete the file data from the storage backend.

        Args:
            fm (FileMetadata): The file metadata

        Returns:
            bool: True if the file was deleted, False otherwise
        """

    def get_public_url(
        self, fm: FileMetadata, expire: Optional[int] = None
    ) -> Optional[str]:
        """Generate a public URL for an existing file.

        Args:
            fm (FileMetadata): The file metadata
            expire (Optional[int], optional): Expiration time in seconds. Defaults to
                class default.

        Returns:
            str: The generated public URL
        """
        return None

    @property
    @abstractmethod
    def save_chunk_size(self) -> int:
        """Get the save chunk size.

        Returns:
            int: The save chunk size
        """


class LocalFileStorage(StorageBackend):
    """Local file storage backend."""

    storage_type: str = "local"

    def __init__(self, base_path: str, save_chunk_size: int = 1024 * 1024):
        """Initialize the local file storage backend."""
        self.base_path = base_path
        self._save_chunk_size = save_chunk_size
        os.makedirs(self.base_path, exist_ok=True)

    @property
    def save_chunk_size(self) -> int:
        """Get the save chunk size."""
        return self._save_chunk_size

    def save(
        self,
        bucket: str,
        file_id: str,
        file_data: BinaryIO,
        public_url: bool = False,
        public_url_expire: Optional[int] = None,
    ) -> str:
        """Save the file data to the local storage backend."""
        bucket_path = os.path.join(self.base_path, bucket)
        os.makedirs(bucket_path, exist_ok=True)
        file_path = os.path.join(bucket_path, file_id)
        with open(file_path, "wb") as f:
            while True:
                chunk = file_data.read(self.save_chunk_size)
                if not chunk:
                    break
                f.write(chunk)
        return file_path

    def load(self, fm: FileMetadata) -> BinaryIO:
        """Load the file data from the local storage backend."""
        bucket_path = os.path.join(self.base_path, fm.bucket)
        file_path = os.path.join(bucket_path, fm.file_id)
        return open(file_path, "rb")  # noqa: SIM115

    def delete(self, fm: FileMetadata) -> bool:
        """Delete the file data from the local storage backend."""
        bucket_path = os.path.join(self.base_path, fm.bucket)
        file_path = os.path.join(bucket_path, fm.file_id)
        if os.path.exists(file_path):
            os.remove(file_path)
            return True
        return False


def calculate_file_hash(file_data: BinaryIO, buffer_size: int) -> str:
    """Calculate the MD5 hash of the file data."""
    hasher = hashlib.md5()
    file_data.seek(0)
    while chunk := file_data.read(buffer_size):
        hasher.update(chunk)
    file_data.seek(0)
    return hasher.hexdigest()


class FileStorageSystem:
    """File storage system."""

    def __init__(
        self,
        storage_backends: Dict[str, StorageBackend],
        metadata_storage: Optional[StorageInterface[FileMetadata, Any]] = None,
        check_hash: bool = True,
    ):
        """Initialize the file storage system."""
        metadata_storage = metadata_storage or InMemoryStorage()
        self.storage_backends = storage_backends
        self.metadata_storage = metadata_storage
        self.check_hash = check_hash
        self._save_chunk_size = min(
            backend.save_chunk_size for backend in storage_backends.values()
        )

    def _calculate_file_hash(self, file_data: BinaryIO) -> str:
        """Calculate the MD5 hash of the file data."""
        if not self.check_hash:
            return "-1"
        return calculate_file_hash(file_data, self._save_chunk_size)

    @trace("file_storage_system.save_file")
    def save_file(
        self,
        bucket: str,
        file_name: str,
        file_data: BinaryIO,
        storage_type: str,
        custom_metadata: Optional[Dict[str, Any]] = None,
        file_id: Optional[str] = None,
        public_url: bool = False,
        public_url_expire: Optional[int] = None,
    ) -> str:
        """Save the file data to the storage backend."""
        file_id = str(uuid.uuid4()) if not file_id else file_id
        backend = self.storage_backends.get(storage_type)
        if not backend:
            raise ValueError(f"Unsupported storage type: {storage_type}")

        with root_tracer.start_span(
            "file_storage_system.save_file.backend_save",
            metadata={
                "bucket": bucket,
                "file_id": file_id,
                "file_name": file_name,
                "storage_type": storage_type,
            },
        ):
            storage_path = backend.save(
                bucket,
                file_id,
                file_data,
                public_url=public_url,
                public_url_expire=public_url_expire,
            )

        file_data.seek(0, 2)  # Move to the end of the file
        file_size = file_data.tell()  # Get the file size
        file_data.seek(0)  # Reset file pointer

        # filter None value
        custom_metadata = (
            {k: v for k, v in custom_metadata.items() if v is not None}
            if custom_metadata
            else {}
        )

        with root_tracer.start_span(
            "file_storage_system.save_file.calculate_hash",
        ):
            file_hash = self._calculate_file_hash(file_data)
        uri = FileStorageURI(
            storage_type, bucket, file_id, custom_params=custom_metadata
        )

        metadata = FileMetadata(
            file_id=file_id,
            bucket=bucket,
            file_name=file_name,
            file_size=file_size,
            storage_type=storage_type,
            storage_path=storage_path,
            uri=str(uri),
            custom_metadata=custom_metadata,
            file_hash=file_hash,
        )

        self.metadata_storage.save(metadata)
        return str(uri)

    @trace("file_storage_system.get_file")
    def get_file(self, uri: str) -> Tuple[BinaryIO, FileMetadata]:
        """Get the file data from the storage backend."""
        if FileStorageURI.is_local_file(uri):
            local_file_name = uri.split("/")[-1]
            if not os.path.exists(uri):
                raise FileNotFoundError(f"File not found: {uri}")

            dummy_metadata = FileMetadata(
                file_id=local_file_name,
                bucket="dummy_bucket",
                file_name=local_file_name,
                file_size=-1,
                storage_type="local",
                storage_path=uri,
                uri=uri,
                custom_metadata={},
                file_hash="",
            )
            logger.info(f"Reading local file: {uri}")
            return open(uri, "rb"), dummy_metadata  # noqa: SIM115

        parsed_uri = FileStorageURI.parse(uri)
        metadata = self.metadata_storage.load(
            FileMetadataIdentifier(
                file_id=parsed_uri.file_id, bucket=parsed_uri.bucket
            ),
            FileMetadata,
        )
        if not metadata:
            raise FileNotFoundError(f"No metadata found for URI: {uri}")

        backend = self.storage_backends.get(metadata.storage_type)
        if not backend:
            raise ValueError(f"Unsupported storage type: {metadata.storage_type}")

        with root_tracer.start_span(
            "file_storage_system.get_file.backend_load",
            metadata={
                "bucket": metadata.bucket,
                "file_id": metadata.file_id,
                "file_name": metadata.file_name,
                "storage_type": metadata.storage_type,
            },
        ):
            file_data = backend.load(metadata)

        with root_tracer.start_span(
            "file_storage_system.get_file.verify_hash",
        ):
            calculated_hash = self._calculate_file_hash(file_data)
        if calculated_hash != "-1" and calculated_hash != metadata.file_hash:
            raise ValueError("File integrity check failed. Hash mismatch.")

        return file_data, metadata

    def get_file_metadata(self, bucket: str, file_id: str) -> Optional[FileMetadata]:
        """Get the file metadata.

        Args:
            bucket (str): The bucket name
            file_id (str): The file ID

        Returns:
            Optional[FileMetadata]: The file metadata
        """
        fid = FileMetadataIdentifier(file_id=file_id, bucket=bucket)
        return self.metadata_storage.load(fid, FileMetadata)

    def get_file_metadata_by_uri(self, uri: str) -> Optional[FileMetadata]:
        """Get the file metadata by URI.

        Args:
            uri (str): The file URI

        Returns:
            Optional[FileMetadata]: The file metadata
        """
        parsed_uri = FileStorageURI.parse(uri)
        return self.get_file_metadata(parsed_uri.bucket, parsed_uri.file_id)

    def delete_file(self, uri: str) -> bool:
        """Delete the file data from the storage backend.

        Args:
            uri (str): The file URI

        Returns:
            bool: True if the file was deleted, False otherwise
        """
        parsed_uri = FileStorageURI.parse(uri)
        fid = FileMetadataIdentifier(
            file_id=parsed_uri.file_id, bucket=parsed_uri.bucket
        )
        metadata = self.metadata_storage.load(fid, FileMetadata)
        if not metadata:
            return False

        backend = self.storage_backends.get(metadata.storage_type)
        if not backend:
            raise ValueError(f"Unsupported storage type: {metadata.storage_type}")

        if backend.delete(metadata):
            try:
                self.metadata_storage.delete(fid)
                return True
            except Exception:
                # If the metadata deletion fails, log the error and return False
                return False
        return False

    def get_public_url(
        self,
        uri: str,
        expire: Optional[int] = None,
    ) -> str:
        """Generate a public URL for an existing file.

        Args:
            uri (str): The file URI
            expire (Optional[int], optional): Expiration time in seconds. Defaults to
                None.

        Returns:
            str: The generated public URL, just return the original URI if the backend
                does not support public URL generation.
        """
        parsed_uri = FileStorageURI.parse(uri)
        metadata = self.metadata_storage.load(
            FileMetadataIdentifier(
                file_id=parsed_uri.file_id, bucket=parsed_uri.bucket
            ),
            FileMetadata,
        )
        if not metadata:
            return None

        backend = self.storage_backends.get(metadata.storage_type)
        if not backend:
            raise ValueError(f"Unsupported storage type: {metadata.storage_type}")

        pub_url = backend.get_public_url(metadata, expire)
        return pub_url if pub_url else uri

    def list_files(
        self, bucket: str, filters: Optional[Dict[str, Any]] = None
    ) -> List[FileMetadata]:
        """List the files in the bucket."""
        filters = filters or {}
        filters["bucket"] = bucket
        return self.metadata_storage.query(QuerySpec(conditions=filters), FileMetadata)


class FileStorageClient(BaseComponent):
    """File storage client component."""

    name = ComponentType.FILE_STORAGE_CLIENT.value

    def __init__(
        self,
        system_app: Optional[SystemApp] = None,
        storage_system: Optional[FileStorageSystem] = None,
        save_chunk_size: int = 1024 * 1024,
        default_storage_type: Optional[str] = None,
    ):
        """Initialize the file storage client."""
        super().__init__(system_app=system_app)
        if not storage_system:
            from pathlib import Path

            base_path = Path.home() / ".cache" / "dbgpt" / "files"
            storage_system = FileStorageSystem(
                {
                    LocalFileStorage.storage_type: LocalFileStorage(
                        base_path=str(base_path)
                    )
                }
            )
        if not default_storage_type:
            if storage_system and storage_system.storage_backends:
                default_storage_type = list(storage_system.storage_backends.keys())[0]

        self.system_app = system_app
        self._storage_system = storage_system
        self.save_chunk_size = save_chunk_size
        self.default_storage_type = default_storage_type

    def init_app(self, system_app: SystemApp):
        """Initialize the application."""
        self.system_app = system_app

    @property
    def storage_system(self) -> FileStorageSystem:
        """Get the file storage system."""
        if not self._storage_system:
            raise ValueError("File storage system not initialized")
        return self._storage_system

    def upload_file(
        self,
        bucket: str,
        file_path: str,
        storage_type: Optional[str] = None,
        custom_metadata: Optional[Dict[str, Any]] = None,
        file_id: Optional[str] = None,
    ) -> str:
        """Upload a file to the storage system.

        Args:
            bucket (str): The bucket name
            file_path (str): The file path
            storage_type (str): The storage type
            custom_metadata (Dict[str, Any], optional): Custom metadata. Defaults to
                None.
            file_id (str, optional): The file ID. Defaults to None. If not provided, a
                random UUID will be generated.

        Returns:
            str: The file URI
        """
        with open(file_path, "rb") as file:
            return self.save_file(
                bucket,
                os.path.basename(file_path),
                file,
                storage_type,
                custom_metadata,
                file_id,
            )

    def save_file(
        self,
        bucket: str,
        file_name: str,
        file_data: BinaryIO,
        storage_type: Optional[str] = None,
        custom_metadata: Optional[Dict[str, Any]] = None,
        file_id: Optional[str] = None,
    ) -> str:
        """Save the file data to the storage system.

        Args:
            bucket (str): The bucket name
            file_name (str): The file name
            file_data (BinaryIO): The file data
            storage_type (str): The storage type
            custom_metadata (Dict[str, Any], optional): Custom metadata. Defaults to
                None.
            file_id(str, optional): The file ID. Defaults to None. If not provided, a
                random UUID will be generated.

        Returns:
            str: The file URI
        """
        if not storage_type:
            storage_type = self.default_storage_type
        if not storage_type:
            raise ValueError("Storage type not provided")
        return self.storage_system.save_file(
            bucket, file_name, file_data, storage_type, custom_metadata, file_id
        )

    def download_file(
        self,
        uri: str,
        dest_path: Optional[str] = None,
        dest_dir: Optional[str] = None,
        cache: bool = True,
    ) -> Tuple[str, FileMetadata]:
        """Download a file from the storage system.

        If dest_path is provided, the file will be saved to that path.
        If dest_dir is provided, the file will be saved to that directory with
        file ID and extension.

        If neither dest_path nor dest_dir is provided, the file will be saved to the
        system temp directory.

        Args:
            uri (str): The file URI
            dest_path (str, optional): The destination path. Defaults to None.
            dest_dir (str, optional): The destination directory. Defaults to None.
            cache (bool, optional): Whether to cache the file. Defaults to True.

        Raises:
            FileNotFoundError: If the file is not found
        """
        file_metadata = self.storage_system.get_file_metadata_by_uri(uri)
        if not file_metadata:
            raise FileNotFoundError(f"File not found: {uri}")

        extension = os.path.splitext(file_metadata.file_name)[1]
        if dest_path:
            target_path = dest_path
        elif dest_dir:
            os.makedirs(dest_dir, exist_ok=True)
            target_path = os.path.join(dest_dir, file_metadata.file_id + extension)
        else:
            from pathlib import Path

            # Write to system temp directory
            base_path = str(Path.home() / ".cache" / "dbgpt" / "files")
            os.makedirs(base_path, exist_ok=True)
            target_path = os.path.join(base_path, file_metadata.file_id + extension)
        file_hash = file_metadata.file_hash
        if os.path.exists(target_path) and cache:
            logger.debug(f"File {target_path} already exists, begin hash check")
            with open(target_path, "rb") as f:
                if file_hash == calculate_file_hash(f, self.save_chunk_size):
                    logger.info(f"File {uri} already exists at {target_path}")
                    return target_path, file_metadata
        logger.info(f"Downloading file {uri} to {target_path}")
        file_data, _ = self.storage_system.get_file(uri)

        with open(target_path, "wb") as f:
            while True:
                chunk = file_data.read(self.save_chunk_size)
                if not chunk:
                    break
                f.write(chunk)
        return target_path, file_metadata

    def get_file(self, uri: str) -> Tuple[BinaryIO, FileMetadata]:
        """Get the file data from the storage system.

        Args:
            uri (str): The file URI

        Returns:
            Tuple[BinaryIO, FileMetadata]: The file data and metadata
        """
        return self.storage_system.get_file(uri)

    def get_file_by_id(
        self, bucket: str, file_id: str
    ) -> Tuple[BinaryIO, FileMetadata]:
        """Get the file data from the storage system by ID.

        Args:
            bucket (str): The bucket name
            file_id (str): The file ID

        Returns:
            Tuple[BinaryIO, FileMetadata]: The file data and metadata
        """
        metadata = self.storage_system.get_file_metadata(bucket, file_id)
        if not metadata:
            raise FileNotFoundError(f"File {file_id} not found in bucket {bucket}")
        return self.get_file(metadata.uri)

    def delete_file(self, uri: str) -> bool:
        """Delete the file data from the storage system.

        Args:
            uri (str): The file URI

        Returns:
            bool: True if the file was deleted, False otherwise
        """
        return self.storage_system.delete_file(uri)

    def delete_file_by_id(self, bucket: str, file_id: str) -> bool:
        """Delete the file data from the storage system by ID.

        Args:
            bucket (str): The bucket name
            file_id (str): The file ID

        Returns:
            bool: True if the file was deleted, False otherwise
        """
        metadata = self.storage_system.get_file_metadata(bucket, file_id)
        if not metadata:
            raise FileNotFoundError(f"File {file_id} not found in bucket {bucket}")
        return self.delete_file(metadata.uri)

    def list_files(
        self, bucket: str, filters: Optional[Dict[str, Any]] = None
    ) -> List[FileMetadata]:
        """List the files in the bucket.

        Args:
            bucket (str): The bucket name
            filters (Dict[str, Any], optional): Filters. Defaults to None.

        Returns:
            List[FileMetadata]: The list of file metadata
        """
        return self.storage_system.list_files(bucket, filters)

    def get_public_url(
        self,
        uri: str,
        expire: Optional[int] = None,
    ) -> str:
        """Generate a public URL for an existing file.

        Args:
            uri (str): The file URI
            expire (Optional[int], optional): Expiration time in seconds. Defaults to
                None.

        Returns:
            str: The generated public URL
        """
        return self.storage_system.get_public_url(uri, expire)


class SimpleDistributedStorage(StorageBackend):
    """Simple distributed storage backend."""

    storage_type: str = "distributed"

    def __init__(
        self,
        node_address: str,
        local_storage_path: str,
        save_chunk_size: int = 1024 * 1024,
        transfer_chunk_size: int = 1024 * 1024,
        transfer_timeout: int = 360,
        api_prefix: str = "/api/v2/serve/file/files",
    ):
        """Initialize the simple distributed storage backend."""
        self.node_address = node_address
        self.local_storage_path = local_storage_path
        os.makedirs(self.local_storage_path, exist_ok=True)
        self._save_chunk_size = save_chunk_size
        self._transfer_chunk_size = transfer_chunk_size
        self._transfer_timeout = transfer_timeout
        self._api_prefix = api_prefix

    @property
    def save_chunk_size(self) -> int:
        """Get the save chunk size."""
        return self._save_chunk_size

    def _get_file_path(self, bucket: str, file_id: str, node_address: str) -> str:
        node_id = hashlib.md5(node_address.encode()).hexdigest()
        return os.path.join(self.local_storage_path, bucket, f"{file_id}_{node_id}")

    def _parse_node_address(self, fm: FileMetadata) -> str:
        storage_path = fm.storage_path
        if not storage_path.startswith("distributed://"):
            raise ValueError("Invalid storage path")
        return storage_path.split("//")[1].split("/")[0]

    def save(
        self,
        bucket: str,
        file_id: str,
        file_data: BinaryIO,
        public_url: bool = False,
        public_url_expire: Optional[int] = None,
    ) -> str:
        """Save the file data to the distributed storage backend.

        Just save the file locally.

        Args:
            bucket (str): The bucket name
            file_id (str): The file ID
            file_data (BinaryIO): The file data

        Returns:
            str: The storage path
        """
        file_path = self._get_file_path(bucket, file_id, self.node_address)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "wb") as f:
            while True:
                chunk = file_data.read(self.save_chunk_size)
                if not chunk:
                    break
                f.write(chunk)

        return f"distributed://{self.node_address}/{bucket}/{file_id}"

    def load(self, fm: FileMetadata) -> BinaryIO:
        """Load the file data from the distributed storage backend.

        If the file is stored on the local node, load it from the local storage.

        Args:
            fm (FileMetadata): The file metadata

        Returns:
            BinaryIO: The file data
        """
        file_id = fm.file_id
        bucket = fm.bucket
        node_address = self._parse_node_address(fm)
        file_path = self._get_file_path(bucket, file_id, node_address)

        # TODO: check if the file is cached in local storage
        if node_address == self.node_address:
            if os.path.exists(file_path):
                return open(file_path, "rb")  # noqa: SIM115
            else:
                raise FileNotFoundError(f"File {file_id} not found on the local node")
        else:
            response = requests.get(
                f"http://{node_address}{self._api_prefix}/{bucket}/{file_id}",
                timeout=self._transfer_timeout,
                stream=True,
            )
            response.raise_for_status()
            # TODO: cache the file in local storage
            return StreamedBytesIO(
                response.iter_content(chunk_size=self._transfer_chunk_size)
            )

    def delete(self, fm: FileMetadata) -> bool:
        """Delete the file data from the distributed storage backend.

        If the file is stored on the local node, delete it from the local storage.
        If the file is stored on a remote node, send a delete request to the remote
        node.

        Args:
            fm (FileMetadata): The file metadata

        Returns:
            bool: True if the file was deleted, False otherwise
        """
        file_id = fm.file_id
        bucket = fm.bucket
        node_address = self._parse_node_address(fm)
        file_path = self._get_file_path(bucket, file_id, node_address)
        if node_address == self.node_address:
            if os.path.exists(file_path):
                os.remove(file_path)
                return True
            return False
        else:
            try:
                response = requests.delete(
                    f"http://{node_address}{self._api_prefix}/{bucket}/{file_id}",
                    timeout=self._transfer_timeout,
                )
                response.raise_for_status()
                return True
            except Exception:
                return False


class StreamedBytesIO(io.BytesIO):
    """A BytesIO subclass that can be used with streaming responses.

    Adapted from: https://gist.github.com/obskyr/b9d4b4223e7eaf4eedcd9defabb34f13
    """

    def __init__(self, request_iterator):
        """Initialize the StreamedBytesIO instance."""
        super().__init__()
        self._bytes = BytesIO()
        self._iterator = request_iterator

    def _load_all(self):
        self._bytes.seek(0, io.SEEK_END)
        for chunk in self._iterator:
            self._bytes.write(chunk)

    def _load_until(self, goal_position):
        current_position = self._bytes.seek(0, io.SEEK_END)
        while current_position < goal_position:
            try:
                current_position += self._bytes.write(next(self._iterator))
            except StopIteration:
                break

    def tell(self) -> int:
        """Get the current position."""
        return self._bytes.tell()

    def read(self, size: Optional[int] = None) -> bytes:
        """Read the data from the stream.

        Args:
            size (Optional[int], optional): The number of bytes to read. Defaults to
                None.

        Returns:
            bytes: The read data
        """
        left_off_at = self._bytes.tell()
        if size is None:
            self._load_all()
        else:
            goal_position = left_off_at + size
            self._load_until(goal_position)

        self._bytes.seek(left_off_at)
        return self._bytes.read(size)

    def seek(self, position: int, whence: int = io.SEEK_SET):
        """Seek to a position in the stream.

        Args:
            position (int): The position
            whence (int, optional): The reference point. Defaults to io.SEEK

        Raises:
            ValueError: If the reference point is invalid
        """
        if whence == io.SEEK_END:
            self._load_all()
        else:
            self._bytes.seek(position, whence)

    def __enter__(self):
        """Enter the context manager."""
        return self

    def __exit__(self, ext_type, value, tb):
        """Exit the context manager."""
        self._bytes.close()



"""Module for summary related classes and functions."""

from .db_summary import (  # noqa: F401
    DBSummary,
    FieldSummary,
    IndexSummary,
    TableSummary,
)

__all__ = ["DBSummary", "FieldSummary", "IndexSummary", "TableSummary"]



"""Graph store base class."""

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass

from dbgpt.util import BaseParameters, RegisterParameters

logger = logging.getLogger(__name__)


@dataclass
class GraphStoreConfig(BaseParameters, RegisterParameters):
    """Graph store config."""

    __cfg_type__ = "graph_store"

    # name: str = Field(
    #     default="dbgpt_collection",
    #     description="The name of graph store, inherit from index store.",
    # )
    # embedding_fn: Optional[Embeddings] = Field(
    #     default=None,
    #     description="The embedding function of graph store, optional.",
    # )
    # enable_summary: bool = Field(
    #     default=False,
    #     description="Enable graph community summary or not.",
    # )
    # enable_similarity_search: bool = Field(
    #     default=False,
    #     description="Enable similarity search or not.",
    # )


class GraphStoreBase(ABC):
    """Graph store base class."""

    def __init__(self, config: GraphStoreConfig):
        """Initialize graph store."""
        self._config = config
        self._conn = None
        self.enable_summary = config.enable_summary
        self.enable_similarity_search = config.enable_similarity_search

    @abstractmethod
    def get_config(self) -> GraphStoreConfig:
        """Get the graph store config."""

    def is_exist(self, name) -> bool:
        """Check Graph Name is Exist."""
        raise NotImplementedError



"""Graph definition."""

import itertools
import json
import logging
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum
from typing import Any, Callable, Dict, Iterator, List, Optional, Set, Tuple

logger = logging.getLogger(__name__)


class GraphElemType(Enum):
    """Type of element in graph."""

    DOCUMENT = "document"
    CHUNK = "chunk"
    ENTITY = "entity"  # default vertex type in knowledge graph
    RELATION = "relation"  # default edge type in knowledge graph
    INCLUDE = "include"
    NEXT = "next"

    DOCUMENT_INCLUDE_CHUNK = "document_include_chunk"
    CHUNK_INCLUDE_CHUNK = "chunk_include_chunk"
    CHUNK_INCLUDE_ENTITY = "chunk_include_entity"
    CHUNK_NEXT_CHUNK = "chunk_next_chunk"

    def is_vertex(self) -> bool:
        """Check if the element is a vertex."""
        return self in [
            GraphElemType.DOCUMENT,
            GraphElemType.CHUNK,
            GraphElemType.ENTITY,
        ]

    def is_edge(self) -> bool:
        """Check if the element is an edge."""
        return self in [
            GraphElemType.RELATION,
            GraphElemType.INCLUDE,
            GraphElemType.NEXT,
            GraphElemType.DOCUMENT_INCLUDE_CHUNK,
            GraphElemType.CHUNK_INCLUDE_CHUNK,
            GraphElemType.CHUNK_INCLUDE_ENTITY,
            GraphElemType.CHUNK_NEXT_CHUNK,
        ]


class Direction(Enum):
    """Direction class."""

    OUT = 0
    IN = 1
    BOTH = 2


class Elem(ABC):
    """Elem class."""

    def __init__(self, name: Optional[str] = None):
        """Initialize Elem."""
        self._name = name
        self._props: Dict[str, Any] = {}

    @property
    def name(self) -> str:
        """Return the edge label."""
        return self._name or ""

    @property
    def props(self) -> Dict[str, Any]:
        """Get all the properties of Elem."""
        return self._props

    def set_prop(self, key: str, value: Any):
        """Set a property of ELem."""
        self._props[key] = value  # note: always update the value

    def get_prop(self, key: str):
        """Get one of the properties of Elem."""
        return self._props.get(key)

    def del_prop(self, key: str):
        """Delete a property of ELem."""
        self._props.pop(key, None)

    def has_props(self, **props):
        """Check all key-value pairs exist."""
        return all(self._props.get(k) == v for k, v in props.items())

    @abstractmethod
    def format(self) -> str:
        """Format properties into a string."""
        if len(self._props) == 1:
            return str(next(iter(self._props.values())))

        formatted_props = [
            f"{k}:{json.dumps(v, ensure_ascii=False)}" for k, v in self._props.items()
        ]
        return f"{{{';'.join(formatted_props)}}}"


class Vertex(Elem):
    """Vertex class."""

    def __init__(self, vid: str, name: Optional[str] = None, **props):
        """Initialize Vertex."""
        super().__init__(name)
        self._vid = vid
        for k, v in props.items():
            self.set_prop(k, v)

    @property
    def vid(self) -> str:
        """Return the vertex ID."""
        return self._vid

    @property
    def name(self) -> str:
        """Return the vertex name."""
        return super().name or self._vid

    def format(self, concise: bool = False):
        """Format vertex into a string."""
        name = self._name or self._vid
        if concise:
            return f"({name})"

        if self._props:
            return f"({name}:{super().format()})"
        else:
            return f"({name})"

    def __str__(self):
        """Return the vertex ID as its string representation."""
        return f"({self._vid})"


class IdVertex(Vertex):
    """IdVertex class."""

    def __init__(self, vid: str):
        """Initialize Idvertex."""
        super().__init__(vid)


class Edge(Elem):
    """Edge class."""

    def __init__(self, sid: str, tid: str, name: str, **props):
        """Initialize Edge."""
        assert name, "Edge name is required"

        super().__init__(name)
        self._sid = sid
        self._tid = tid
        for k, v in props.items():
            self.set_prop(k, v)

    def __eq__(self, other):
        """Check if two edges are equal.

        Let's say two edges are equal if they have the same source vertex ID,
        target vertex ID, and edge label. The properties are not considered.
        """
        return (self.sid, self.tid, self.name) == (other.sid, other.tid, other.name)

    def __hash__(self):
        """Return the hash value of the edge."""
        return hash((self.sid, self.tid, self.name))

    @property
    def sid(self) -> str:
        """Return the source vertex ID of the edge."""
        return self._sid

    @property
    def tid(self) -> str:
        """Return the target vertex ID of the edge."""
        return self._tid

    def nid(self, vid):
        """Return neighbor id."""
        if vid == self._sid:
            return self._tid
        elif vid == self._tid:
            return self._sid
        else:
            raise ValueError(f"Get nid of {vid} on {self} failed")

    def format(self):
        """Format the edge properties into a string."""
        if self._props:
            return f"-[{self._name}:{super().format()}]->"
        else:
            return f"-[{self._name}]->"

    def triplet(self) -> Tuple[str, str, str]:
        """Return a triplet."""
        return self.sid, self.name, self.tid

    def __str__(self):
        """Return the edge '(sid)->(tid)'."""
        return f"({self._sid})-[{self._name}]->({self._tid})"


class Graph(ABC):
    """Graph class."""

    @abstractmethod
    def upsert_vertex(self, vertex: Vertex):
        """Add a vertex."""

    @abstractmethod
    def append_edge(self, edge: Edge):
        """Add an edge."""

    @abstractmethod
    def has_vertex(self, vid: str) -> bool:
        """Check vertex exists."""

    @abstractmethod
    def get_vertex(self, vid: str) -> Vertex:
        """Get a vertex."""

    @abstractmethod
    def get_neighbor_edges(
        self,
        vid: str,
        direction: Direction = Direction.OUT,
        limit: Optional[int] = None,
    ) -> Iterator[Edge]:
        """Get neighbor edges."""

    @abstractmethod
    def vertices(
        self, filter_fn: Optional[Callable[[Vertex], bool]] = None
    ) -> Iterator[Vertex]:
        """Get vertex iterator."""

    @abstractmethod
    def edges(
        self, filter_fn: Optional[Callable[[Edge], bool]] = None
    ) -> Iterator[Edge]:
        """Get edge iterator."""

    @abstractmethod
    def del_vertices(self, *vids: str):
        """Delete vertices and their neighbor edges."""

    @abstractmethod
    def del_edges(self, sid: str, tid: str, name: str, **props):
        """Delete edges(sid -[name]-> tid) matches props."""

    @abstractmethod
    def del_neighbor_edges(self, vid: str, direction: Direction = Direction.OUT):
        """Delete neighbor edges."""

    @abstractmethod
    def search(
        self,
        vids: List[str],
        direct: Direction = Direction.OUT,
        depth: Optional[int] = None,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> "Graph":
        """Search on graph."""

    @abstractmethod
    def schema(self) -> Dict[str, Any]:
        """Get schema."""

    @abstractmethod
    def format(self) -> str:
        """Format graph data to string."""

    @abstractmethod
    def truncate(self):
        """Truncate graph."""


class MemoryGraph(Graph):
    """Graph class."""

    def __init__(self):
        """Initialize MemoryGraph with vertex label and edge label."""
        # metadata
        self._vertex_prop_keys = set()
        self._edge_prop_keys = set()
        self._edge_count = 0

        # vertices index, out edges index, in edges index
        self._vs: Any = defaultdict()
        self._oes: Any = defaultdict(lambda: defaultdict(set))
        self._ies: Any = defaultdict(lambda: defaultdict(set))

    @property
    def vertex_count(self):
        """Return the number of vertices in the graph."""
        return len(self._vs)

    @property
    def edge_count(self):
        """Return the count of edges in the graph."""
        return self._edge_count

    def upsert_vertex(self, vertex: Vertex):
        """Insert or update a vertex based on its ID."""
        if vertex.vid in self._vs:
            if isinstance(self._vs[vertex.vid], IdVertex):
                self._vs[vertex.vid] = vertex
            else:
                self._vs[vertex.vid].props.update(vertex.props)
        else:
            self._vs[vertex.vid] = vertex

        # update metadata
        self._vertex_prop_keys.update(vertex.props.keys())

    def append_edge(self, edge: Edge) -> bool:
        """Append an edge if it doesn't exist; requires edge label."""
        sid = edge.sid
        tid = edge.tid

        if edge in self._oes[sid][tid]:
            return False

        # init vertex index
        self._vs.setdefault(sid, IdVertex(sid))
        self._vs.setdefault(tid, IdVertex(tid))

        # update edge index
        self._oes[sid][tid].add(edge)
        self._ies[tid][sid].add(edge)

        # update metadata
        self._edge_prop_keys.update(edge.props.keys())
        self._edge_count += 1
        return True

    def upsert_graph(self, graph: "MemoryGraph"):
        """Upsert a graph."""
        for vertex in graph.vertices():
            self.upsert_vertex(vertex)

        for edge in graph.edges():
            self.append_edge(edge)

    def has_vertex(self, vid: str) -> bool:
        """Retrieve a vertex by ID."""
        return vid in self._vs

    def get_vertex(self, vid: str) -> Vertex:
        """Retrieve a vertex by ID."""
        return self._vs[vid]

    def get_neighbor_edges(
        self,
        vid: str,
        direction: Direction = Direction.OUT,
        limit: Optional[int] = None,
    ) -> Iterator[Edge]:
        """Get edges connected to a vertex by direction."""
        if direction == Direction.OUT:
            es = (e for es in self._oes[vid].values() for e in es)

        elif direction == Direction.IN:
            es = iter(e for es in self._ies[vid].values() for e in es)

        elif direction == Direction.BOTH:
            oes = (e for es in self._oes[vid].values() for e in es)
            ies = (e for es in self._ies[vid].values() for e in es)

            # merge
            tuples = itertools.zip_longest(oes, ies)
            es = (e for t in tuples for e in t if e is not None)

            # distinct
            seen = set()

            # es = (e for e in es if e not in seen and not seen.add(e))
            def unique_elements(elements):
                for element in elements:
                    if element not in seen:
                        seen.add(element)
                        yield element

            es = unique_elements(es)
        else:
            raise ValueError(f"Invalid direction: {direction}")

        return itertools.islice(es, limit) if limit else es

    def vertices(
        self, filter_fn: Optional[Callable[[Vertex], bool]] = None
    ) -> Iterator[Vertex]:
        """Return vertices."""
        # Get all vertices in the graph
        all_vertices = self._vs.values()

        return all_vertices if filter_fn is None else filter(filter_fn, all_vertices)

    def edges(
        self, filter_fn: Optional[Callable[[Edge], bool]] = None
    ) -> Iterator[Edge]:
        """Return edges."""
        # Get all edges in the graph
        all_edges = (e for nbs in self._oes.values() for es in nbs.values() for e in es)

        if filter_fn is None:
            return all_edges
        else:
            return filter(filter_fn, all_edges)

    def del_vertices(self, *vids: str):
        """Delete specified vertices."""
        for vid in vids:
            self.del_neighbor_edges(vid, Direction.BOTH)
            self._vs.pop(vid, None)

    def del_edges(self, sid: str, tid: str, name: str, **props):
        """Delete edges."""
        old_edge_cnt = len(self._oes[sid][tid])

        def remove_matches(es: Set[Edge]):
            return set(
                filter(
                    lambda e: (
                        not (
                            (name == e.name if name else True) and e.has_props(**props)
                        )
                    ),
                    es,
                )
            )

        self._oes[sid][tid] = remove_matches(self._oes[sid][tid])
        self._ies[tid][sid] = remove_matches(self._ies[tid][sid])

        self._edge_count -= old_edge_cnt - len(self._oes[sid][tid])

    def del_neighbor_edges(self, vid: str, direction: Direction = Direction.OUT):
        """Delete all neighbor edges."""

        def del_index(idx, i_idx):
            for nid in idx[vid].keys():
                self._edge_count -= len(i_idx[nid][vid])
                i_idx[nid].pop(vid, None)
            idx.pop(vid, None)

        if direction in [Direction.OUT, Direction.BOTH]:
            del_index(self._oes, self._ies)

        if direction in [Direction.IN, Direction.BOTH]:
            del_index(self._ies, self._oes)

    def search(
        self,
        vids: List[str],
        direct: Direction = Direction.OUT,
        depth: Optional[int] = None,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> "MemoryGraph":
        """Search the graph from a vertex with specified parameters."""
        subgraph = MemoryGraph()

        for vid in vids:
            self.__search(vid, direct, depth, fan, limit, 0, set(), subgraph)

        return subgraph

    def __search(
        self,
        vid: str,
        direct: Direction,
        depth: Optional[int],
        fan: Optional[int],
        limit: Optional[int],
        _depth: int,
        _visited: Set,
        _subgraph: "MemoryGraph",
    ):
        if vid in _visited or depth and _depth >= depth:
            return

        # visit vertex
        if not self.has_vertex(vid):
            return
        _subgraph.upsert_vertex(self.get_vertex(vid))
        _visited.add(vid)

        # visit edges
        nids = set()
        for edge in self.get_neighbor_edges(vid, direct, fan):
            if limit and _subgraph.edge_count >= limit:
                return

            # append edge success then visit new vertex
            if _subgraph.append_edge(edge):
                nid = edge.nid(vid)
                if nid not in _visited:
                    nids.add(nid)

        # next hop
        for nid in nids:
            self.__search(
                nid, direct, depth, fan, limit, _depth + 1, _visited, _subgraph
            )

    def schema(self) -> Dict[str, Any]:
        """Return schema."""
        return {
            "schema": [
                {
                    "type": "VERTEX",
                    "properties": [{"name": k} for k in self._vertex_prop_keys],
                },
                {
                    "type": "EDGE",
                    "properties": [{"name": k} for k in self._edge_prop_keys],
                },
            ]
        }

    def format(self, entities_only: Optional[bool] = False) -> str:
        """Format graph to string."""
        vs_str = "\n".join(v.format() for v in self.vertices())
        es_str = "\n".join(
            f"{self.get_vertex(e.sid).format(concise=True)}"
            f"{e.format()}"
            f"{self.get_vertex(e.tid).format(concise=True)}"
            for e in self.edges()
        )
        if entities_only:
            return f"Entities:\n{vs_str}" if vs_str else ""
        else:
            return (
                f"Entities:\n{vs_str}\n\nRelationships:\n{es_str}"
                if (vs_str or es_str)
                else ""
            )

    def truncate(self):
        """Truncate graph."""
        # clean metadata
        self._vertex_prop_keys.clear()
        self._edge_prop_keys.clear()
        self._edge_count = 0

        # clean data and index
        self._vs.clear()
        self._oes.clear()
        self._ies.clear()

    def graphviz(self, name="g"):
        """View graphviz graph: https://dreampuf.github.io/GraphvizOnline."""
        try:
            import networkx as nx
        except ImportError:
            raise ImportError(
                "networkx is required for graph visualization, please install it by "
                "running `pip install networkx`"
            )

        g = nx.MultiDiGraph()
        for vertex in self.vertices():
            g.add_node(vertex.vid)

        for edge in self.edges():
            triplet = edge.triplet()
            g.add_edge(triplet[0], triplet[2], label=triplet[1])

        digraph = nx.nx_agraph.to_agraph(g).to_string()
        digraph = digraph.replace('digraph ""', f"digraph {name}")
        digraph = re.sub(r"key=\d+,?\s*", "", digraph)
        return digraph



"""Hybrid memory module.

This structure explicitly models the human short-term and long-term memories. The
short-term memory temporarily buffers recent perceptions, while long-term memory
consolidates important information over time.
"""

import os.path
from concurrent.futures import Executor, ThreadPoolExecutor
from datetime import datetime
from typing import Generic, List, Optional, Tuple, Type

from dbgpt.core import Embeddings, LLMClient
from dbgpt.storage.vector_store.base import VectorStoreBase
from dbgpt.util.annotations import immutable, mutable

from .base import (
    DiscardedMemoryFragments,
    ImportanceScorer,
    InsightExtractor,
    Memory,
    SensoryMemory,
    ShortTermMemory,
    T,
    WriteOperation,
)
from .long_term import LongTermMemory
from .short_term import EnhancedShortTermMemory


class HybridMemory(Memory, Generic[T]):
    """Hybrid memory for the agent."""

    importance_weight: float = 0.9

    def __init__(
        self,
        now: datetime,
        sensory_memory: SensoryMemory[T],
        short_term_memory: ShortTermMemory[T],
        long_term_memory: LongTermMemory[T],
        default_insight_extractor: Optional[InsightExtractor] = None,
        default_importance_scorer: Optional[ImportanceScorer] = None,
    ):
        """Create a hybrid memory."""
        self.now = now
        self._sensory_memory = sensory_memory
        self._short_term_memory = short_term_memory
        self._long_term_memory = long_term_memory
        self._default_insight_extractor = default_insight_extractor
        self._default_importance_scorer = default_importance_scorer

    def structure_clone(
        self: "HybridMemory[T]", now: Optional[datetime] = None
    ) -> "HybridMemory[T]":
        """Return a structure clone of the memory."""
        now = now or self.now
        m = HybridMemory(
            now=now,
            sensory_memory=self._sensory_memory.structure_clone(now),
            short_term_memory=self._short_term_memory.structure_clone(now),
            long_term_memory=self._long_term_memory.structure_clone(now),
            default_insight_extractor=self._default_insight_extractor,
            default_importance_scorer=self._default_importance_scorer,
        )
        m._copy_from(self)
        return m

    @classmethod
    def from_chroma(
        cls,
        vstore_name: Optional[str] = "agent_memory_long_term",
        vstore_path: Optional[str] = None,
        embeddings: Optional[Embeddings] = None,
        executor: Optional[Executor] = None,
        now: Optional[datetime] = None,
        sensory_memory: Optional[SensoryMemory[T]] = None,
        short_term_memory: Optional[ShortTermMemory[T]] = None,
        long_term_memory: Optional[LongTermMemory[T]] = None,
        **kwargs,
    ):
        """Create a hybrid memory from Chroma vector store."""
        from dbgpt.configs.model_config import DATA_DIR
        from dbgpt_ext.storage.vector_store.chroma_store import (
            ChromaStore,
            ChromaVectorConfig,
        )

        if not embeddings:
            from dbgpt.rag.embedding import DefaultEmbeddingFactory

            embeddings = DefaultEmbeddingFactory.openai()

        vstore_path = vstore_path or os.path.join(DATA_DIR, "agent_memory")

        vector_store = ChromaStore(
            ChromaVectorConfig(persist_path=vstore_path),
            name=vstore_name,
            embedding_fn=embeddings,
        )
        return cls.from_vstore(
            vector_store=vector_store,
            embeddings=embeddings,
            executor=executor,
            now=now,
            sensory_memory=sensory_memory,
            short_term_memory=short_term_memory,
            long_term_memory=long_term_memory,
            **kwargs,
        )

    @classmethod
    def from_vstore(
        cls,
        vector_store: "VectorStoreBase",
        embeddings: Optional[Embeddings] = None,
        executor: Optional[Executor] = None,
        now: Optional[datetime] = None,
        sensory_memory: Optional[SensoryMemory[T]] = None,
        short_term_memory: Optional[ShortTermMemory[T]] = None,
        long_term_memory: Optional[LongTermMemory[T]] = None,
        **kwargs,
    ):
        """Create a hybrid memory from vector store."""
        if not embeddings:
            raise ValueError("embeddings is required.")
        if not executor:
            executor = ThreadPoolExecutor()
        if not now:
            now = datetime.now()

        if not sensory_memory:
            sensory_memory = SensoryMemory()
        if not short_term_memory:
            if not embeddings:
                raise ValueError("embeddings is required.")
            short_term_memory = EnhancedShortTermMemory(embeddings, executor)
        if not long_term_memory:
            long_term_memory = LongTermMemory(
                executor,
                vector_store,
                now=now,
            )
        return cls(now, sensory_memory, short_term_memory, long_term_memory, **kwargs)

    def initialize(
        self,
        name: Optional[str] = None,
        llm_client: Optional[LLMClient] = None,
        importance_scorer: Optional[ImportanceScorer[T]] = None,
        insight_extractor: Optional[InsightExtractor[T]] = None,
        real_memory_fragment_class: Optional[Type[T]] = None,
        session_id: Optional[str] = None,
    ) -> None:
        """Initialize the memory.

        It will initialize all the memories.
        """
        memories = [
            self._sensory_memory,
            self._short_term_memory,
            self._long_term_memory,
        ]
        kwargs = {
            "name": name,
            "llm_client": llm_client,
            "importance_scorer": importance_scorer,
            "insight_extractor": insight_extractor,
            "real_memory_fragment_class": real_memory_fragment_class,
            "session_id": session_id,
        }
        for memory in memories:
            memory.initialize(**kwargs)
        super().initialize(**kwargs)

    @mutable
    async def write(
        self,
        memory_fragment: T,
        now: Optional[datetime] = None,
        op: WriteOperation = WriteOperation.ADD,
    ) -> Optional[DiscardedMemoryFragments[T]]:
        """Write a memory fragment to the memory."""
        return await self._write_single(
            memory_fragment,
            now=now,
            op=op,
            write_long_term=True,
        )

    async def _write_single(
        self,
        memory_fragment: T,
        now: Optional[datetime] = None,
        op: WriteOperation = WriteOperation.ADD,
        write_long_term: bool = True,
    ) -> Optional[DiscardedMemoryFragments[T]]:
        """Write a single memory fragment to the memory."""
        # First write to sensory memory
        sen_discarded_memories = await self._sensory_memory.write(
            memory_fragment, now=now, op=op
        )
        if not sen_discarded_memories:
            return None
        short_term_discarded_memories = []
        discarded_memory_fragments = []
        discarded_insights = []
        for sen_memory in sen_discarded_memories.discarded_memory_fragments:
            # Write to short term memory
            short_discarded_memory = await self._short_term_memory.write(
                sen_memory, now=now, op=op
            )
            if short_discarded_memory:
                short_term_discarded_memories.append(short_discarded_memory)
                discarded_memory_fragments.extend(
                    short_discarded_memory.discarded_memory_fragments
                )
                for insight in short_discarded_memory.discarded_insights:
                    # Just keep the first insight
                    discarded_insights.append(insight.insights[0])
        if not write_long_term:
            return None
        # Obtain the importance of insights
        insight_scores = await self.score_memory_importance(discarded_insights)
        # Get the importance of insights
        for i, ins in enumerate(discarded_insights):
            ins.update_importance(insight_scores[i])
        all_memories = discarded_memory_fragments + discarded_insights
        if self._long_term_memory and len(all_memories) > 0:
            # Write to long term memory
            await self._long_term_memory.write_batch(all_memories, self.now)
        return None

    @mutable
    async def write_batch(
        self, memory_fragments: List[T], now: Optional[datetime] = None
    ) -> Optional[DiscardedMemoryFragments[T]]:
        """Write a batch of memory fragments to the memory.

        For memory recovery, we only write to sensory memory and short term memory.
        """
        for memory_fragment in memory_fragments:
            # Just write to sensory memory and short term memory
            await self._write_single(
                memory_fragment,
                now=now,
                op=WriteOperation.ADD,
                write_long_term=False,
            )

    @immutable
    async def read(
        self,
        observation: str,
        alpha: Optional[float] = None,
        beta: Optional[float] = None,
        gamma: Optional[float] = None,
    ) -> List[T]:
        """Read memories from the memory."""
        (
            retrieved_long_term_memories,
            short_term_discarded_memories,
        ) = await self.fetch_memories(observation, self._short_term_memory)

        await self.save_memories_after_retrieval(short_term_discarded_memories)
        return retrieved_long_term_memories

    @immutable
    async def fetch_memories(
        self,
        observation: str,
        short_term_memory: Optional[ShortTermMemory[T]] = None,
    ) -> Tuple[List[T], List[DiscardedMemoryFragments[T]]]:
        """Fetch memories from long term memory.

        If short_term_memory is provided, write the fetched memories to the short term
        memory.
        """
        retrieved_long_term_memories = await self._long_term_memory.fetch_memories(
            observation
        )
        if not short_term_memory:
            return retrieved_long_term_memories, []
        short_term_discarded_memories: List[DiscardedMemoryFragments[T]] = []
        discarded_memory_fragments: List[T] = []
        for ltm in retrieved_long_term_memories:
            short_discarded_memory = await short_term_memory.write(
                ltm, op=WriteOperation.RETRIEVAL
            )
            if short_discarded_memory:
                short_term_discarded_memories.append(short_discarded_memory)
                discarded_memory_fragments.extend(
                    short_discarded_memory.discarded_memory_fragments
                )
        for stm in short_term_memory.short_term_memories:
            retrieved_long_term_memories.append(
                stm.current_class.build_from(
                    observation=stm.raw_observation,
                    importance=stm.importance,
                )
            )
        return retrieved_long_term_memories, short_term_discarded_memories

    async def save_memories_after_retrieval(
        self, fragments: List[DiscardedMemoryFragments[T]]
    ):
        """Save memories after retrieval."""
        discarded_memory_fragments = []
        discarded_memory_insights: List[T] = []
        for f in fragments:
            discarded_memory_fragments.extend(f.discarded_memory_fragments)
            for fi in f.discarded_insights:
                discarded_memory_insights.append(fi.insights[0])
        insights_importance = await self.score_memory_importance(
            discarded_memory_insights
        )
        for i, ins in enumerate(discarded_memory_insights):
            ins.update_importance(insights_importance[i])
        all_memories = discarded_memory_fragments + discarded_memory_insights
        await self._long_term_memory.write_batch(all_memories, self.now)

    async def clear(self) -> List[T]:
        """Clear the memory.

        # TODO
        """
        return []



"""Extended types for the OpenAI API."""

from typing_extensions import Literal, Required, TypedDict


class ExtAudioURL(TypedDict, total=False):
    url: Required[str]
    """URL of the audio file."""

    detail: Required[Literal["wav", "mp3"]]
    """The format of the encoded audio data."""


class ExtChatCompletionContentPartInputAudioParam(TypedDict, total=False):
    audio_url: Required[ExtAudioURL]

    type: Required[Literal["audio_url"]]
    """The type of the content part. Always `audio_url`."""


class ExtVideoURL(TypedDict, total=False):
    url: Required[str]
    """URL of the video file."""

    detail: Required[Literal["mp4", "avi"]]
    """The format of the encoded video data."""


class ExtChatCompletionContentPartInputVideoParam(TypedDict, total=False):
    video_url: Required[ExtVideoURL]

    type: Required[Literal["video_url"]]
    """The type of the content part. Always `video_url`."""


class ExtFileURL(TypedDict, total=False):
    url: Required[str]
    """URL of the file."""


class ExtChatCompletionContentPartInputFileParam(TypedDict, total=False):
    file_url: Required[ExtFileURL]
    type: Required[Literal["file_url"]]
    """The type of the content part. Always `file_url`."""
